# pytest: ollama, llm

from mellea import generative, start_session
from mellea.backends import ModelOption
from mellea.core import CBlock
from mellea.stdlib.context import ChatContext

# Generative slots can be used with sessions that have context.
# By utilizing context, you can change the results of several
# functions at the same time. This lets you write a function
# once but use it in different ways.


@generative
def grade_essay(essay: str) -> int:
    """Grades the provided essay.

    Args:
        essay: the text to be graded.

    Returns:
        int: a grade between 1 and 100.
    """


@generative
def give_feedback(essay: str) -> list[str]:
    """Generates feedback for improvement for a given essay.

    Args:
        essay: the text that should be commented on.

    Returns:
        list[str]: a list of comments about how the text could be improved.
    """


if __name__ == "__main__":
    m = start_session(
        ctx=ChatContext(), model_options={ModelOption.MAX_NEW_TOKENS: 100}
    )

    text = """
"Mabble" is an obsolete verb meaning to wrap up.
Its only recorded usage is in the early 1600s.
Mabble is a variant of the word "moble", a transitive verb meaning
to muffle (a person, or the head, face, etc...) figuratively
or literally.
"""

    # Generative functions can be used as is, without additional context.
    print("Giving a grade and comments on the essay with no context:")
    grade = grade_essay(m, essay=text)
    print(f"  Grade: {grade}")

    comments = give_feedback(m, essay=text)
    print(f"  Comments: {' '.join(comments)}")

    # If you have a set of generative functions, you can tweak them all by
    # adding context to the session they are running in.
    m.ctx = m.ctx.add(
        CBlock(
            "You are an elementary school teacher. "
            "Any grades and feedback that you give should keep that in mind. Remember to be "
            "especially kind and considerate so that you don't hurt the students' feelings. "
            "Try to keep all grades above an 86 unless something is very wrong."
        )
    )
    print(
        "\n\n\nGiving a grade and comments on the essay with elementary school context:"
    )
    grade = grade_essay(m, essay=text)
    print(f"  Grade: {grade}")

    comments = give_feedback(m, essay=text)
    print(f"  Comments: {' '.join(comments)}")

    # And, let's reset the context and try a different grading style.
    m.reset()
    m.ctx = m.ctx.add(
        CBlock(
            "You are a grammarian that is focused solely on spelling and syntax, "
            "not on the content of essays. When giving grades and feedback, focus "
            "on spelling errors, contractions, and other grammar issues."
        )
    )
    print("\n\n\nGiving a grade and comments on the essay with grammar context:")
    grade = grade_essay(m, essay=text)
    print(f"  Grade: {grade}")

    comments = give_feedback(m, essay=text)
    print(f"  Comments: {' '.join(comments)}")



import re
from collections.abc import Callable
from typing import Any, TypeVar, final

from mellea import MelleaSession
from mellea.backends import ModelOption
from mellea.stdlib.components import Message

from .._prompt_modules import PromptModule, PromptModuleString
from ._exceptions import BackendGenerationError, TagExtractionError
from ._prompt import get_system_prompt, get_user_prompt

T = TypeVar("T")

RE_GENERAL_INSTRUCTIONS = re.compile(
    r"<general_instructions>(.+?)</general_instructions>",
    flags=re.IGNORECASE | re.DOTALL,
)


@final
class _GeneralInstructions(PromptModule):
    @staticmethod
    def _default_parser(generated_str: str) -> str:
        general_instructions_match = re.search(RE_GENERAL_INSTRUCTIONS, generated_str)

        general_instructions_str: str | None = (
            general_instructions_match.group(1).strip()
            if general_instructions_match
            else None
        )

        if general_instructions_str is None:
            raise TagExtractionError(
                'LLM failed to generate correct tags for extraction: "<general_instructions>"'
            )

        return general_instructions_str

    def generate(
        self,
        mellea_session: MelleaSession,
        input_str: str | None,
        max_new_tokens: int = 4096,
        parser: Callable[[str], T] = _default_parser,  # type: ignore[assignment]
        # About the mypy ignore above: https://github.com/python/mypy/issues/3737
        **kwargs: dict[str, Any],
    ) -> PromptModuleString[T]:
        assert input_str is not None, 'This module requires the "input_str" argument'

        system_prompt = get_system_prompt()
        user_prompt = get_user_prompt(task_prompt=input_str)

        action = Message("user", user_prompt)

        try:
            gen_result = mellea_session.act(
                action=action,
                model_options={
                    ModelOption.SYSTEM_PROMPT: system_prompt,
                    ModelOption.TEMPERATURE: 0,
                    ModelOption.MAX_NEW_TOKENS: max_new_tokens,
                },
            ).value
        except Exception as e:
            raise BackendGenerationError(f"LLM generation failed: {e}")

        if gen_result is None:
            raise BackendGenerationError(
                "LLM generation failed: value attribute is None"
            )

        return PromptModuleString(gen_result, parser)


general_instructions = _GeneralInstructions()



"""``ChatFormatter`` for converting context histories to chat-message lists.

``ChatFormatter`` is the standard formatter used by mellea's legacy backends. Its
``to_chat_messages`` method linearises a sequence of ``Component`` and ``CBlock``
objects into ``Message`` objects with ``user``, ``assistant``, or ``tool`` roles,
handling ``ModelOutputThunk`` responses, image attachments, and parsed structured
outputs. Concrete backends call this formatter when preparing input for a chat
completion endpoint.
"""

from ..core import (
    CBlock,
    Component,
    Formatter,
    ModelOutputThunk,
    TemplateRepresentation,
)
from ..stdlib.components.chat import Message


class ChatFormatter(Formatter):
    """Formatter used by Legacy backends to format Contexts as Messages."""

    def to_chat_messages(self, cs: list[Component | CBlock]) -> list[Message]:
        """Convert a linearized chat history into a list of chat messages.

        Iterates over each element in the context history and converts it to a
        ``Message`` with an appropriate role. ``ModelOutputThunk`` instances are
        treated as assistant responses, while all other ``Component`` and
        ``CBlock`` objects default to the ``user`` role. Image attachments and
        parsed structured outputs are handled transparently.

        Args:
            cs (list[Component | CBlock]): The linearized sequence of context
                components and code blocks to convert.

        Returns:
            list[Message]: A list of ``Message`` objects ready for submission to
                a chat completion endpoint.
        """

        def _to_msg(c: Component | CBlock) -> Message:
            role: Message.Role = "user"  # default to `user`; see ModelOutputThunk below for when the role changes.

            # Check if it's a ModelOutputThunk first since that changes what we should be printing
            # as the message content.
            if isinstance(c, ModelOutputThunk):
                role = "assistant"  # ModelOutputThunks should always be responses from a model.

                assert c.is_computed()
                assert (
                    c.value is not None
                )  # This is already entailed by c.is_computed(); the line is included here to satisfy the type-checker.

                if c.parsed_repr is not None:
                    if isinstance(c.parsed_repr, Component):
                        # Only use the parsed_repr if it's something that we know how to print.
                        c = c.parsed_repr  # This might be a message.
                    else:
                        # Otherwise, explicitly stringify it.
                        c = Message(role=role, content=str(c.parsed_repr))
                else:
                    c = Message(role=role, content=c.value)  # type: ignore

            match c:
                case Message():
                    return c
                case Component():
                    images = None
                    tr = c.format_for_llm()
                    if isinstance(tr, TemplateRepresentation):
                        images = tr.images

                    # components can have images
                    return Message(role=role, content=self.print(c), images=images)
                case _:
                    return Message(role=role, content=self.print(c))

        return [_to_msg(c) for c in cs]



"""``MObject``, ``Query``, ``Transform``, and ``MObjectProtocol`` for query/transform workflows.

Defines the ``MObjectProtocol`` protocol for objects that can be queried and
transformed by an LLM, and the concrete ``MObject`` base class that implements it.
Also provides the ``Query`` and ``Transform`` ``Component`` subtypes, which wrap an
object with a natural-language question or mutation instruction respectively. These
primitives underpin ``@mify`` and can be composed directly to build document Q&A
or structured extraction pipelines.
"""

from __future__ import annotations

import inspect
from collections.abc import Callable
from typing import Protocol, runtime_checkable

from ...backends.tools import MelleaTool
from ...core import CBlock, Component, ModelOutputThunk, TemplateRepresentation


class Query(Component[str]):
    """A ``Component`` that pairs an ``MObject`` with a natural-language question.

    Wraps the object and its query string into a ``TemplateRepresentation`` so the
    formatter can render both together in a prompt, optionally forwarding the
    object's tools and fields to the template.

    Args:
        obj (Component): The object to be queried.
        query (str): The natural-language question to ask about the object.
    """

    def __init__(self, obj: Component, query: str) -> None:
        """Initialize Query with the object to query and a natural-language question string."""
        self._obj = obj
        self._query = query

    def parts(self) -> list[Component | CBlock]:
        """Return the constituent parts of this query component.

        Returns:
            list[Component | CBlock]: A list containing the wrapped object.
        """
        return [self._obj]

    def format_for_llm(self) -> TemplateRepresentation | str:
        """Format this query for the language model.

        Returns:
            TemplateRepresentation | str: A ``TemplateRepresentation`` containing
            the query string, the wrapped object, and any tools or fields from the
            object's own representation.
        """
        object_repr = self._obj.format_for_llm()
        return TemplateRepresentation(
            args={
                "query": self._query,
                "content": self._obj,  # Put the object here so the object template can be applied first.
            },
            obj=self,
            tools=(
                object_repr.tools
                if isinstance(object_repr, TemplateRepresentation)
                else None
            ),
            fields=(
                object_repr.fields
                if isinstance(object_repr, TemplateRepresentation)
                else None
            ),
            template_order=["Query"],
        )

    def _parse(self, computed: ModelOutputThunk) -> str:
        """Parse the model output. Returns string value for now."""
        return computed.value if computed.value is not None else ""


class Transform(Component[str]):
    """A ``Component`` that pairs an ``MObject`` with a natural-language mutation instruction.

    Wraps the object and its transformation description into a
    ``TemplateRepresentation`` so the formatter can render both together in a prompt,
    optionally forwarding the object's tools and fields to the template.

    Args:
        obj (Component): The object to be transformed.
        transformation (str): The natural-language description of the transformation.
    """

    def __init__(self, obj: Component, transformation: str) -> None:
        """Initialize Transform with the object to transform and a natural-language description."""
        self._obj = obj
        self._transformation = transformation

    def parts(self) -> list[Component | CBlock]:
        """Return the constituent parts of this transform component.

        Returns:
            list[Component | CBlock]: A list containing the wrapped object.
        """
        return [self._obj]

    def format_for_llm(self) -> TemplateRepresentation | str:
        """Format this transform for the language model.

        Returns:
            TemplateRepresentation | str: A ``TemplateRepresentation`` containing
            the transformation description, the wrapped object, and any tools or
            fields from the object's own representation.
        """
        object_repr = self._obj.format_for_llm()
        return TemplateRepresentation(
            args={
                "transformation": self._transformation,
                "content": self._obj,  # Put the object here so the object template can be applied first.
            },
            obj=self,
            tools=(
                object_repr.tools
                if isinstance(object_repr, TemplateRepresentation)
                else None
            ),
            fields=(
                object_repr.fields
                if isinstance(object_repr, TemplateRepresentation)
                else None
            ),
            template_order=["Transform"],
        )

    def _parse(self, computed: ModelOutputThunk) -> str:
        """Parse the model output. Returns string value for now."""
        return computed.value if computed.value is not None else ""


@runtime_checkable
class MObjectProtocol(Protocol):
    """Protocol to describe the necessary functionality of a MObject. Implementers should prefer inheriting from MObject than MObjectProtocol."""

    def parts(self) -> list[Component | CBlock]:
        """Return a list of parts for this MObject.

        Returns:
            list[Component | CBlock]: The constituent sub-components.
        """
        ...

    def get_query_object(self, query: str) -> Query:
        """Return the instantiated query object.

        Args:
            query (str): The query string.

        Returns:
            Query: A ``Query`` component wrapping this object and the given
            query string.
        """
        ...

    def get_transform_object(self, transformation: str) -> Transform:
        """Return the instantiated transform object.

        Args:
            transformation (str): The transformation description string.

        Returns:
            Transform: A ``Transform`` component wrapping this object and the
            given transformation description.
        """
        ...

    def content_as_string(self) -> str:
        """Return the content of this MObject as a plain string.

        The default value is just ``str(self)``.
        Subclasses should override this method.

        Returns:
            str: String representation of this object's content.
        """
        ...

    def _get_all_members(self) -> dict[str, Callable]:
        """Return all methods from this MObject that are not inherited from the superclass.

        Undocumented methods and methods with ``[no-index]`` in their docstring
        are ignored.
        """
        ...

    def format_for_llm(self) -> TemplateRepresentation | str:
        """Return the template representation used by the formatter.

        The default ``TemplateRepresentation`` uses automatic parsing for tools
        and fields. Content is retrieved from ``content_as_string()``.

        Returns:
            TemplateRepresentation | str: The formatted representation for the
            language model.
        """
        ...

    def _parse(self, computed: ModelOutputThunk) -> str:
        """Parse the model output."""
        ...


class MObject(Component[str]):
    """An extension of ``Component`` for adding query and transform operations.

    Args:
        query_type (type): The ``Query`` subclass to use when constructing query
            components. Defaults to ``Query``.
        transform_type (type): The ``Transform`` subclass to use when constructing
            transform components. Defaults to ``Transform``.
    """

    def __init__(
        self, *, query_type: type = Query, transform_type: type = Transform
    ) -> None:
        """Initialize MObject with a query type and transform type for building query/transform components."""
        self._query_type = query_type
        self._transform_type = transform_type

    def parts(self) -> list[Component | CBlock]:
        """MObject has no parts because of how format_for_llm is defined.

        Returns:
            list[Component | CBlock]: Always an empty list.
        """
        return []

    def get_query_object(self, query: str) -> Query:
        """Return the instantiated query object.

        Args:
            query (str): The query string.

        Returns:
            Query: A ``Query`` component wrapping this object and the given
            query string.
        """
        return self._query_type(self, query)

    def get_transform_object(self, transformation: str) -> Transform:
        """Return the instantiated transform object.

        Args:
            transformation (str): The transformation description string.

        Returns:
            Transform: A ``Transform`` component wrapping this object and the
            given transformation description.
        """
        return self._transform_type(self, transformation)

    def content_as_string(self) -> str:
        """Return the content of this MObject as a plain string.

        The default value is just ``str(self)``.
        Subclasses should override this method.

        Returns:
            str: String representation of this object's content.
        """
        return str(self)

    def _get_all_members(self) -> dict[str, Callable]:
        """Return all methods from this MObject except methods of the superclass.

        Undocumented methods and methods with ``[no-index]`` in their docstring
        are ignored.
        """
        all_members: dict[str, Callable] = dict(
            inspect.getmembers(self, predicate=inspect.ismethod)
        )
        unique_members = {}

        # Get members of superclass
        superclass_members = dict(inspect.getmembers(MObject)).keys()

        # Filter out members that are also in superclasses
        for name, member in all_members.items():
            if name not in superclass_members and (
                hasattr(member, "__doc__")
                and member.__doc__ is not None
                and "[no-index]" not in member.__doc__.strip()
            ):
                unique_members[name] = member
        return unique_members

    def format_for_llm(self) -> TemplateRepresentation | str:
        """Return the template representation used by the formatter.

        The default ``TemplateRepresentation`` uses automatic parsing for tools
        and fields. Content is retrieved from ``content_as_string()``.

        Returns:
            TemplateRepresentation | str: The formatted representation for the
            language model.
        """
        tools = {
            k: MelleaTool.from_callable(c) for k, c in self._get_all_members().items()
        }
        return TemplateRepresentation(
            args={"content": self.content_as_string()},
            obj=self,
            tools=tools,
            fields=[],
            template_order=["*", "MObject"],
        )

    def _parse(self, computed: ModelOutputThunk) -> str:
        """Parse the model output. Returns string value for now."""
        return computed.value if computed.value is not None else ""



# pytest: ollama, llm

from textwrap import fill
from typing import Any

from mellea.core import Requirement, ValidationResult


# Just for printing stuff nicely...
def w(x: Any) -> str:
    return fill(str(x), width=120, replace_whitespace=False)


def req_print(rv_list: list[tuple[Requirement, ValidationResult]]) -> str:
    parts = [f"{bool(rv[1])}\t: {rv[0].description}" for rv in rv_list]
    return "\n".join(parts)



"""Sampling pipeline hook payloads."""

from __future__ import annotations

from typing import Any

from mellea.plugins.base import MelleaBasePayload


class SamplingLoopStartPayload(MelleaBasePayload):
    """Payload for ``sampling_loop_start`` — when sampling strategy begins.

    Attributes:
        strategy_name: Class name of the sampling strategy (e.g. ``"RejectionSamplingStrategy"``).
        action: The ``Component`` being sampled.

        context: The ``Context`` at the start of sampling.

        requirements: List of ``Requirement`` instances to validate against.
        loop_budget: Maximum number of sampling iterations allowed (writable).
    """

    strategy_name: str = ""
    action: Any = None
    context: Any = None
    requirements: list[Any] = []
    loop_budget: int = 0


class SamplingIterationPayload(MelleaBasePayload):
    """Payload for ``sampling_iteration`` — after each sampling attempt.

    Attributes:
        iteration: 1-based iteration number within the sampling loop.
        action: The ``Component`` used for this attempt.

        result: The ``ModelOutputThunk`` produced by this attempt.
        validation_results: List of ``(Requirement, ValidationResult)`` tuples.
        all_validations_passed: ``True`` when **every** requirement in ``validation_results``
            passed for this iteration (i.e., the sampling attempt succeeded).
        valid_count: Number of requirements that passed.
        total_count: Total number of requirements evaluated.
    """

    iteration: int = 0
    action: Any = None
    result: Any = None
    validation_results: list[tuple[Any, Any]] = []
    all_validations_passed: bool = False
    valid_count: int = 0
    total_count: int = 0


class SamplingRepairPayload(MelleaBasePayload):
    """Payload for ``sampling_repair`` — when repair is invoked after validation failure.

    Attributes:
        repair_type: Kind of repair (strategy-dependent, e.g. ``"rejection"``, ``"template"``).
        failed_action: The ``Component`` that failed validation.

        failed_result: The ``ModelOutputThunk`` that failed validation.
        failed_validations: List of ``(Requirement, ValidationResult)`` tuples that failed.
        repair_action: The repaired ``Component`` to use for the next attempt.
        repair_context: The ``Context`` to use for the next attempt.
        repair_iteration: 1-based iteration at which the repair was triggered.
    """

    repair_type: str = ""
    failed_action: Any = None
    failed_result: Any = None
    failed_validations: list[tuple[Any, Any]] = []
    repair_action: Any = None
    repair_context: Any = None
    repair_iteration: int = 0


class SamplingLoopEndPayload(MelleaBasePayload):
    """Payload for ``sampling_loop_end`` — when sampling completes.

    Attributes:
        success: ``True`` if at least one attempt passed all requirements.
        iterations_used: Total number of iterations the loop executed.
        final_result: The selected ``ModelOutputThunk`` (best success or best failure).
        final_action: The ``Component`` that produced ``final_result``.

        final_context: The ``Context`` associated with ``final_result``.

        failure_reason: Human-readable reason when ``success`` is ``False``.
        all_results: List of ``ModelOutputThunk`` from every iteration.
        all_validations: Nested list — ``all_validations[i]`` is the list of
            ``(Requirement, ValidationResult)`` tuples for iteration *i*.
    """

    success: bool = False
    iterations_used: int = 0
    final_result: Any = None
    final_action: Any = None
    final_context: Any = None
    failure_reason: str | None = None
    all_results: list[Any] = []
    all_validations: list[list[tuple[Any, Any]]] = []



"""A model backend wrapping the Ollama Python SDK."""

import asyncio
import datetime
import functools
from collections.abc import AsyncIterator, Callable, Coroutine, Sequence
from typing import Any, overload

import ollama
from tqdm import tqdm

from ..backends import ModelIdentifier, model_ids
from ..core import (
    BaseModelSubclass,
    C,
    CBlock,
    Component,
    Context,
    FancyLogger,
    GenerateLog,
    GenerateType,
    ModelOutputThunk,
    ModelToolCall,
)
from ..core.base import AbstractMelleaTool
from ..formatters import ChatFormatter, TemplateFormatter
from ..helpers import ClientCache, get_current_event_loop, send_to_queue
from ..stdlib.components import Message
from ..stdlib.requirements import ALoraRequirement
from ..telemetry.backend_instrumentation import (
    instrument_generate_from_context,
    instrument_generate_from_raw,
)
from .backend import FormatterBackend
from .model_options import ModelOption
from .tools import add_tools_from_context_actions, add_tools_from_model_options

format: None = None  # typing this variable in order to shadow the global format function and ensure mypy checks for errors


class OllamaModelBackend(FormatterBackend):
    """A model that uses the Ollama Python SDK for local inference.

    Args:
        model_id (str | ModelIdentifier): Ollama model ID. If a
            ``ModelIdentifier`` is passed, its ``ollama_name`` attribute must
            be set.
        formatter (ChatFormatter | None): Formatter for rendering components.
            Defaults to ``TemplateFormatter``.
        base_url (str | None): Ollama server endpoint; defaults to
            ``env(OLLAMA_HOST)`` or ``http://localhost:11434``.
        model_options (dict | None): Default model options for generation requests.

    Attributes:
        to_mellea_model_opts_map (dict): Mapping from Ollama-specific option names
            to Mellea ``ModelOption`` sentinel keys.
        from_mellea_model_opts_map (dict): Mapping from Mellea ``ModelOption``
            sentinel keys to Ollama-specific option names.
    """

    def __init__(
        self,
        model_id: str | ModelIdentifier = model_ids.IBM_GRANITE_4_MICRO_3B,
        formatter: ChatFormatter | None = None,
        base_url: str | None = None,
        model_options: dict | None = None,
    ):
        """Initialize an Ollama backend, connecting to the server and pulling the model if needed."""
        super().__init__(
            model_id=model_id,
            formatter=(
                formatter
                if formatter is not None
                else TemplateFormatter(model_id=model_id)
            ),
            model_options=model_options,
        )
        # Run the ollama model id accessor early, so that an Assertion fails immediately if we cannot find an ollama model id for the provided ModelIdentifier.
        self._get_ollama_model_id()

        # Setup the client and ensure that we have the model available.
        self._base_url = base_url
        self._client = ollama.Client(base_url)

        self._client_cache = ClientCache(2)

        # Call once to set up an async client and prepopulate the cache.
        _ = self._async_client

        if not self._check_ollama_server():
            err = f"could not create OllamaModelBackend: ollama server not running at {base_url}"
            FancyLogger.get_logger().error(err)
            raise Exception(err)
        if not self._pull_ollama_model():
            err = f"could not create OllamaModelBackend: {self._get_ollama_model_id()} could not be pulled from ollama library"
            FancyLogger.get_logger().error(err)
            raise Exception(err)

        # A mapping of common options for this backend mapped to their Mellea ModelOptions equivalent.
        # These are usually values that must be extracted before hand or that are common among backend providers.
        self.to_mellea_model_opts_map = {
            "system": ModelOption.SYSTEM_PROMPT,
            "think": ModelOption.THINKING,
            "num_ctx": ModelOption.CONTEXT_WINDOW,
            "num_predict": ModelOption.MAX_NEW_TOKENS,
            "seed": ModelOption.SEED,
            "tools": ModelOption.TOOLS,
            "stream": ModelOption.STREAM,
        }

        # A mapping of Mellea specific ModelOptions to the specific names for this backend.
        # These options should almost always be a subset of those specified in the `to_mellea_model_opts_map`.
        # Usually, values that are intentionally extracted while prepping for the backend generate call
        # will be omitted here so that they will be removed when model_options are processed
        # for the call to the model.
        self.from_mellea_model_opts_map = {
            ModelOption.CONTEXT_WINDOW: "num_ctx",
            ModelOption.MAX_NEW_TOKENS: "num_predict",
            ModelOption.SEED: "seed",
        }

    def _get_ollama_model_id(self) -> str:
        """Gets the ollama model id from the model_id that was provided in the constructor. Raises AssertionError is the ModelIdentifier does not provide an ollama_name."""
        ollama_model_id = (
            self.model_id.ollama_name
            if isinstance(self.model_id, ModelIdentifier)
            else self.model_id
        )
        assert ollama_model_id is not None, (
            "model_id is None. This can also happen if the ModelIdentifier has no ollama name set or this model is not available in ollama."
        )
        return ollama_model_id

    def _check_ollama_server(self) -> bool:
        """Requests generic info about the Ollama server to ensure it's running."""
        try:
            self._client.ps()
        except ConnectionError:
            return False
        return True

    def is_model_available(self, model_name):
        """Checks if a specific Ollama model is available locally.

        Args:
          model_name: The name of the model to check for (e.g., "llama2").

        Returns:
          True if the model is available, False otherwise.
        """
        try:
            models = self._client.list()
            for model in models["models"]:
                if model.model.startswith(model_name):
                    return True
            return False
        except Exception as e:
            print(f"An error occurred: {e}")
            return False

    def _pull_ollama_model(self) -> bool:
        """Either gets the cached ollama model or else attempts to pull the provided model from Ollama. Raises an exception of the model cannot be pulled.

        This code was generated by ChatGPT.
        """
        # shortcut --  if model is in list-- don't try to pull
        if self.is_model_available(self._get_ollama_model_id()):
            return True

        try:
            FancyLogger.get_logger().debug(
                f"Loading/Pulling model from Ollama: {self._get_ollama_model_id()}"
            )
            stream = self._client.pull(self._get_ollama_model_id(), stream=True)
            progress_bars = {}
            for update in stream:
                status = update.status
                digest = update.digest
                completed = update.completed or 0
                total = update.total or 0
                # Only track digests with a known total
                if digest and total > 0:
                    if digest not in progress_bars:
                        progress_bars[digest] = tqdm(
                            total=total,
                            desc=f"{status} {digest[:12]}",
                            unit="B",
                            unit_scale=True,
                            leave=False,
                        )
                    pbar = progress_bars[digest]
                    delta = completed - pbar.n
                    if delta > 0:
                        pbar.update(delta)
            # Close all progress bars
            for pbar in progress_bars.values():
                pbar.close()
            return True
        except ollama.ResponseError:
            return False

    @property
    def _async_client(self) -> ollama.AsyncClient:
        """Ollama's client gets tied to a specific event loop. Reset it if needed here."""
        key = id(get_current_event_loop())

        _async_client = self._client_cache.get(key)
        if _async_client is None:
            _async_client = ollama.AsyncClient(self._base_url)
            self._client_cache.put(key, _async_client)
        return _async_client

    def _simplify_and_merge(
        self, model_options: dict[str, Any] | None
    ) -> dict[str, Any]:
        """Simplifies model_options to use the Mellea specific ModelOption.Option and merges the backend's model_options with those passed into this call.

        Rules:
        - Within a model_options dict, existing keys take precedence. This means remapping to mellea specific keys will maintain the value of the mellea specific key if one already exists.
        - When merging, the keys/values from the dictionary passed into this function take precedence.

        Because this function simplifies and then merges, non-Mellea keys from the passed in model_options will replace
        Mellea specific keys from the backend's model_options.

        Args:
            model_options: the model_options for this call

        Returns:
            a new dict
        """
        backend_model_opts = ModelOption.replace_keys(
            self.model_options, self.to_mellea_model_opts_map
        )

        if model_options is None:
            return backend_model_opts

        generate_call_model_opts = ModelOption.replace_keys(
            model_options, self.to_mellea_model_opts_map
        )
        return ModelOption.merge_model_options(
            backend_model_opts, generate_call_model_opts
        )

    def _make_backend_specific_and_remove(
        self, model_options: dict[str, Any]
    ) -> dict[str, Any]:
        """Maps specified Mellea specific keys to their backend specific version and removes any remaining Mellea keys.

        Args:
            model_options: the model_options for this call

        Returns:
            a new dict
        """
        backend_specific = ModelOption.replace_keys(
            model_options, self.from_mellea_model_opts_map
        )
        return ModelOption.remove_special_keys(backend_specific)

    async def _generate_from_context(
        self,
        action: Component[C] | CBlock,
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> tuple[ModelOutputThunk[C], Context]:
        """Generate a completion for ``action`` given ``ctx`` via the Ollama chat API.

        Delegates to ``generate_from_chat_context``. Only chat contexts are supported.

        Args:
            action (Component[C] | CBlock): The component or content block to generate
                a completion for.
            ctx (Context): The current generation context (must be a chat context).
            format (type[BaseModelSubclass] | None): Optional Pydantic model class for
                structured/constrained output decoding.
            model_options (dict | None): Per-call model options that override the
                backend's defaults.
            tool_calls (bool): If ``True``, expose available tools to the model and
                parse tool-call responses.

        Returns:
            tuple[ModelOutputThunk[C], Context]: A thunk holding the (lazy) model output
                and an updated context that includes ``action`` and the new output.
        """
        from ..telemetry.backend_instrumentation import start_generate_span

        # Start span without auto-closing (will be closed in post_processing)
        span = start_generate_span(self, action, ctx, format, tool_calls)

        assert ctx.is_chat_context, (
            "The ollama backend only supports chat-like contexts."
        )
        mot = await self.generate_from_chat_context(
            action,
            ctx,
            _format=format,
            model_options=model_options,
            tool_calls=tool_calls,
        )

        # Store span for telemetry recording and closing in post_processing
        if span is not None:
            mot._meta["_telemetry_span"] = span

        return mot, ctx.add(action).add(mot)

    async def generate_from_chat_context(
        self,
        action: Component[C] | CBlock,
        ctx: Context,
        *,
        _format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> ModelOutputThunk[C]:
        """Generate a new completion from the provided context using this backend's formatter.

        Treats the ``Context`` as a chat history and uses the ``ollama.Client.chat()``
        interface to generate a completion. Returns a thunk that lazily resolves
        the model output.

        Args:
            action (Component[C] | CBlock): The component or content block to generate
                a completion for.
            ctx (Context): The current generation context (must be a chat context).
            _format (type[BaseModelSubclass] | None): Optional Pydantic model class for
                structured output decoding.
            model_options (dict | None): Per-call model options.
            tool_calls (bool): If ``True``, expose available tools and parse responses.

        Returns:
            ModelOutputThunk[C]: A thunk holding the (lazy) model output.

        Raises:
            RuntimeError: If not called from a thread with a running event loop.
        """
        # Start by awaiting any necessary computation.
        await self.do_generate_walk(action)

        model_opts = self._simplify_and_merge(model_options)

        linearized_context = ctx.view_for_generation()
        assert linearized_context is not None, (
            "Cannot generate from a non-linear context in a FormatterBackend."
        )
        # Convert our linearized context into a sequence of chat messages. Template formatters have a standard way of doing this.
        messages: list[Message] = self.formatter.to_chat_messages(linearized_context)
        # Add the final message.
        match action:
            case ALoraRequirement():
                raise Exception(
                    "The ollama backend does not support currently support activated LoRAs."
                )
            case _:
                messages.extend(self.formatter.to_chat_messages([action]))
        # construct the conversation from our messages, adding a system prompt at the first message if one was provided.
        conversation: list[dict] = []
        # We use system prompt None/empty-string semantics in a way that is consistent with huggingface and other libraries.
        # If the system prompt is None, the the default system prompt gets used.
        system_prompt = model_opts.get(ModelOption.SYSTEM_PROMPT, "")
        if system_prompt != "":
            conversation.append({"role": "system", "content": system_prompt})

        conversation.extend(
            [
                {"role": m.role, "content": m.content, "images": m.images}
                for m in messages
            ]
        )

        # Append tool call information if applicable.
        tools: dict[str, AbstractMelleaTool] = dict()
        if tool_calls:
            if _format:
                FancyLogger.get_logger().warning(
                    f"Tool calling typically uses constrained generation, but you have specified a `format` in your generate call. NB: tool calling is superseded by format; we will NOT call tools for your request: {action}"
                )
            else:
                add_tools_from_model_options(tools, model_opts)
                add_tools_from_context_actions(tools, ctx.actions_for_available_tools())

                # Add the tools from the action for this generation last so that
                # they overwrite conflicting names.
                add_tools_from_context_actions(tools, [action])
            FancyLogger.get_logger().info(f"Tools for call: {tools.keys()}")

        # Generate a chat response from ollama, using the chat messages. Can be either type since stream is passed as a model option.
        chat_response: Coroutine[
            Any, Any, AsyncIterator[ollama.ChatResponse] | ollama.ChatResponse
        ] = self._async_client.chat(
            model=self._get_ollama_model_id(),
            messages=conversation,
            tools=[t.as_json_tool for t in tools.values()],
            think=model_opts.get(ModelOption.THINKING, None),
            stream=model_opts.get(ModelOption.STREAM, False),
            options=self._make_backend_specific_and_remove(model_opts),
            format=_format.model_json_schema() if _format is not None else None,  # type: ignore
        )  # type: ignore

        output = ModelOutputThunk(None)
        output._start = datetime.datetime.now()
        output._context = linearized_context
        output._action = action
        output._model_options = model_opts

        # Processing functions only pass the ModelOutputThunk (and current chunk of response). Bind the other vars necessary for
        # each processing step.
        output._process = functools.partial(self.processing, tools=tools)
        output._post_process = functools.partial(
            self.post_processing,
            conversation=conversation,
            tools=tools,
            _format=_format,
        )

        try:
            # To support lazy computation, will need to remove this create_task and store just the unexecuted coroutine.
            # We can also support synchronous calls by adding a flag and changing this ._generate function.

            # This function should always be called from a running event loop so we don't have to worry about
            # scheduling the task to a specific event loop here.

            # Use `create_task` so that we don't have to specifically await this task before it starts executing.
            output._generate = asyncio.create_task(
                send_to_queue(chat_response, output._async_queue)
            )
            output._generate_type = GenerateType.ASYNC
        except RuntimeError as e:
            # Most likely cause is running this function without an event loop present
            raise e

        return output

    @overload
    async def generate_from_raw(
        self,
        actions: list[Component[C]],
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> list[ModelOutputThunk[C]]: ...

    @overload
    async def generate_from_raw(
        self,
        actions: list[Component[C] | CBlock],
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> list[ModelOutputThunk[C | str]]: ...

    async def generate_from_raw(
        self,
        actions: Sequence[Component[C] | CBlock],
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> list[ModelOutputThunk]:
        """Generate completions for multiple actions without chat templating via Ollama.

        Passes formatted prompt strings directly to Ollama's generate endpoint.
        Requests are submitted concurrently to make use of Ollama's concurrency support.

        Args:
            actions (Sequence[Component[C] | CBlock]): Actions to generate completions for.
            ctx (Context): The current generation context.
            format (type[BaseModelSubclass] | None): Optional Pydantic model for
                structured output decoding.
            model_options (dict | None): Per-call model options.
            tool_calls (bool): Ignored; tool calling is not supported on this endpoint.

        Returns:
            list[ModelOutputThunk]: A list of model output thunks, one per action.
        """
        if len(actions) > 1:
            FancyLogger.get_logger().info(
                "Ollama doesn't support batching; will attempt to process concurrently."
            )
        if tool_calls:
            FancyLogger.get_logger().warning(
                "The completion endpoint does not support tool calling at the moment."
            )

        model_opts = self._simplify_and_merge(model_options)

        await self.do_generate_walks(list(actions))
        prompts = [self.formatter.print(action) for action in actions]

        # Ollama doesn't support "batching". There's some ability for concurrency. Use that here.
        # See https://github.com/ollama/ollama/blob/main/docs/faq.md#how-does-ollama-handle-concurrent-requests.

        with instrument_generate_from_raw(
            backend=self, num_actions=len(actions), format=format, tool_calls=tool_calls
        ):
            # Run async so that we can make use of Ollama's concurrency.
            coroutines: list[Coroutine[Any, Any, ollama.GenerateResponse]] = []
            for prompt in prompts:
                co = self._async_client.generate(
                    model=self._get_ollama_model_id(),
                    prompt=prompt,
                    raw=True,
                    think=model_opts.get(ModelOption.THINKING, None),
                    format=format.model_json_schema() if format is not None else None,  # type: ignore
                    options=self._make_backend_specific_and_remove(model_opts),
                )
                coroutines.append(co)

            responses = await asyncio.gather(*coroutines, return_exceptions=True)

        results = []
        date = datetime.datetime.now()
        for i, response in enumerate(responses):
            result = None
            error = None
            if isinstance(response, BaseException):
                FancyLogger.get_logger().warning(
                    f"generate_from_raw: request {i} failed with "
                    f"{type(response).__name__}: {response}"
                )
                result = ModelOutputThunk(value="")
                error = response
            else:
                result = ModelOutputThunk(
                    value=response.response,
                    meta={
                        "generate_response": response.model_dump(),
                        "usage": {
                            "completion_tokens": response.eval_count,
                            "prompt_tokens": response.prompt_eval_count,
                            "total_tokens": (
                                response.prompt_eval_count + response.eval_count
                                if response.prompt_eval_count is not None
                                and response.eval_count is not None
                                else None
                            ),
                        },
                    },
                )
            action = actions[i]
            result.parsed_repr = (
                action.parse(result) if isinstance(action, Component) else result.value
            )

            generate_log = GenerateLog()
            generate_log.prompt = prompts[i]
            generate_log.backend = f"ollama::{self.model_id!s}"
            generate_log.date = date
            generate_log.model_options = model_opts
            generate_log.model_output = result.value
            generate_log.extra = {
                "format": format,
                "thinking": model_opts.get(ModelOption.THINKING, None),
                "seed": model_opts.get(ModelOption.SEED, None),
            }
            generate_log.action = action

            if error:
                generate_log.extra["error"] = error
            result._generate_log = generate_log

            results.append(result)

        return results

    def _extract_model_tool_requests(
        self, tools: dict[str, AbstractMelleaTool], chat_response: ollama.ChatResponse
    ) -> dict[str, ModelToolCall] | None:
        from .tools import validate_tool_arguments

        model_tool_calls: dict[str, ModelToolCall] = {}

        if chat_response.message.tool_calls:
            for tool in chat_response.message.tool_calls:
                func = tools.get(tool.function.name)
                if func is None:
                    FancyLogger.get_logger().warning(
                        f"model attempted to call a non-existing function: {tool.function.name}"
                    )
                    continue  # skip this function if we can't find it.

                args = tool.function.arguments

                # Validate and coerce argument types
                validated_args = validate_tool_arguments(func, args, strict=False)
                model_tool_calls[tool.function.name] = ModelToolCall(
                    tool.function.name, func, validated_args
                )

        if len(model_tool_calls) > 0:
            return model_tool_calls
        return None

    async def processing(
        self,
        mot: ModelOutputThunk,
        chunk: ollama.ChatResponse,
        tools: dict[str, AbstractMelleaTool],
    ):
        """Accumulate text and tool calls from a single Ollama ChatResponse chunk.

        Called for each streaming or non-streaming ``ollama.ChatResponse``. Also
        extracts tool call requests inline and merges the chunk into the running
        aggregated response stored in ``mot._meta["chat_response"]``.

        Args:
            mot (ModelOutputThunk): The output thunk being populated.
            chunk (ollama.ChatResponse): A single chat response object from Ollama.
            tools (dict[str, AbstractMelleaTool]): Available tools, keyed by name,
                used for extracting tool call requests from the response.
        """
        if mot._thinking is None:
            mot._thinking = ""
        thinking_chunk = chunk.message.thinking
        if thinking_chunk is not None:
            mot._thinking += thinking_chunk

        if mot._underlying_value is None:
            mot._underlying_value = ""
        content_chunk = chunk.message.content
        if content_chunk is not None:
            mot._underlying_value += content_chunk

        tool_chunk = self._extract_model_tool_requests(tools, chunk)
        if tool_chunk is not None:
            # Only set tool_calls if there is one.
            if mot.tool_calls is None:
                mot.tool_calls = {}

            # Merge the tool_chunk dict.
            for key, val in tool_chunk.items():
                mot.tool_calls[key] = val

        # Ollama responses are mostly self-contained. Merge chunks immediately.
        chat_response_delta_merge(mot, chunk)

    async def post_processing(
        self,
        mot: ModelOutputThunk,
        conversation: list[dict],
        tools: dict[str, AbstractMelleaTool],
        _format,
    ):
        """Finalize the output thunk after Ollama generation completes.

        Attaches the generate log, records token usage metrics, emits telemetry,
        and cleans up the span reference.

        Args:
            mot (ModelOutputThunk): The output thunk to finalize.
            conversation (list[dict]): The chat conversation sent to the model,
                used for logging.
            tools (dict[str, AbstractMelleaTool]): Available tools, keyed by name.
            _format: The structured output format class used during generation, if any.
        """
        assert mot._action is not None, (
            "ModelOutputThunks should have their action assigned during generation"
        )
        assert mot._model_options is not None, (
            "ModelOutputThunks should have their model_opts assigned during generation"
        )

        # Generate the log for this ModelOutputThunk.
        generate_log = GenerateLog()
        generate_log.prompt = conversation
        generate_log.backend = f"ollama::{self._get_ollama_model_id()}"
        generate_log.model_options = mot._model_options
        generate_log.date = datetime.datetime.now()
        generate_log.model_output = mot._meta["chat_response"]
        generate_log.extra = {
            "format": _format,
            "thinking": mot._model_options.get(ModelOption.THINKING, None),
            "tools_available": tools,
            "tools_called": mot.tool_calls,
            "seed": mot._model_options.get(ModelOption.SEED, None),
        }
        generate_log.action = mot._action
        generate_log.result = mot

        mot._generate_log = generate_log
        mot._generate = None

        # Extract token counts from response
        response = mot._meta.get("chat_response")
        prompt_tokens = (
            getattr(response, "prompt_eval_count", None) if response else None
        )
        completion_tokens = getattr(response, "eval_count", None) if response else None

        # Populate standardized usage field (convert to OpenAI format)
        if prompt_tokens is not None and completion_tokens is not None:
            mot.usage = {
                "prompt_tokens": prompt_tokens,
                "completion_tokens": completion_tokens,
                "total_tokens": prompt_tokens + completion_tokens,
            }

        # Populate model and provider metadata
        mot.model = (
            self.model_id.ollama_name
            if isinstance(self.model_id, ModelIdentifier)
            else str(self.model_id)
        )
        mot.provider = "ollama"

        # Record telemetry and close span now that response is available
        span = mot._meta.get("_telemetry_span")
        if span is not None:
            from ..telemetry import end_backend_span
            from ..telemetry.backend_instrumentation import (
                record_response_metadata,
                record_token_usage,
            )

            if response:
                if mot.usage:
                    record_token_usage(span, mot.usage)
                record_response_metadata(span, response)

            # Close the span now that telemetry is recorded
            end_backend_span(span)

            # Clean up the span reference
            del mot._meta["_telemetry_span"]


def chat_response_delta_merge(mot: ModelOutputThunk, delta: ollama.ChatResponse):
    """Merges the individual ChatResponse chunks from a streaming response into a single ChatResponse.

    Args:
        mot: the ModelOutputThunk that the deltas are being used to populated.
        delta: the most recent ollama ChatResponse.
    """
    if mot._meta.get("chat_response", None) is None:
        mot._meta["chat_response"] = delta
        return  # Return early, no need to merge.

    merged: ollama.ChatResponse = mot._meta["chat_response"]
    if not merged.done:
        merged.done = delta.done
    if merged.done_reason is None:
        merged.done_reason = delta.done_reason
    if merged.total_duration is None:
        merged.total_duration = delta.total_duration
    if merged.load_duration is None:
        merged.load_duration = delta.load_duration
    if merged.prompt_eval_count is None:
        merged.prompt_eval_count = delta.prompt_eval_count
    if merged.prompt_eval_duration is None:
        merged.prompt_eval_duration = delta.prompt_eval_duration
    if merged.eval_count is None:
        merged.eval_count = delta.eval_count

    if merged.message.role == "":
        merged.message.role = delta.message.role

    if merged.message.content is None:
        merged.message.content = delta.message.content
    elif delta.message.content is not None:
        merged.message.content += delta.message.content

    if merged.message.thinking is None:
        merged.message.thinking = delta.message.thinking
    elif delta.message.thinking is not None:
        merged.message.thinking += delta.message.thinking

    if merged.message.tool_calls is None:
        merged.message.tool_calls = delta.message.tool_calls
    elif delta.message.tool_calls is not None:
        merged.message.tool_calls = [
            *merged.message.tool_calls,
            *delta.message.tool_calls,
        ]



# pytest: ollama, llm

from docs.examples.helper import w
from mellea import start_session
from mellea.backends import ModelOption

# create a session using Granite 4 Micro 3B on Ollama and a simple context [see below]
m = start_session(model_options={ModelOption.MAX_NEW_TOKENS: 200})

# Write a more formal and a more funny email
email_v1 = m.instruct("Write an email to invite all interns to the office party.")
email_v2 = m.instruct(
    "Write a very funny email to invite all interns to the office party."
)
print(
    f"***** email 1 ****\n{w(email_v1)}\n*******email 2 ******\n{w(email_v2)}\n\n*******"
)

# Use the emails as grounding context to evaluate which one is quirkier
email_checking = m.instruct(
    "Which email is more quirky?",
    grounding_context={"email_1": email_v1, "email_2": email_v2},
)
print(w(email_checking))



"""Concrete ``Context`` implementations for common conversation patterns.

Provides ``ChatContext``, which accumulates all turns in a sliding-window chat history
(configurable via ``window_size``), and ``SimpleContext``, in which each interaction
is treated as a stateless single-turn exchange (no prior history is passed to the
model). Import ``ChatContext`` for multi-turn conversations and ``SimpleContext`` when
you want each call to the model to be independent.
"""

from __future__ import annotations

# Leave unused `ContextTurn` import for import ergonomics.
from ..core import CBlock, Component, Context, ContextTurn


class ChatContext(Context):
    """Initializes a chat context with unbounded window_size and is_chat=True by default.

    Args:
        window_size (int | None): Maximum number of context turns to include when
            calling ``view_for_generation``. ``None`` (the default) means the full
            history is always returned.
    """

    def __init__(self, *, window_size: int | None = None):
        """Initialize ChatContext with an optional sliding-window size."""
        super().__init__()
        self._window_size = window_size

    def add(self, c: Component | CBlock) -> ChatContext:
        """Add a new component or CBlock to the context and return the updated context.

        Args:
            c (Component | CBlock): The component or content block to append.

        Returns:
            ChatContext: A new ``ChatContext`` with the added entry, preserving the
            current ``window_size`` setting.
        """
        new = ChatContext.from_previous(self, c)
        new._window_size = self._window_size
        return new

    def view_for_generation(self) -> list[Component | CBlock] | None:
        """Return the context entries to pass to the model, respecting the configured window.

        Uses the ``window_size`` set during initialisation to limit how many past
        turns are included.  ``None`` is returned when the underlying history is
        non-linear.

        Returns:
            list[Component | CBlock] | None: Ordered list of context entries up to
            ``window_size`` turns, or ``None`` if the history is non-linear.
        """
        return self.as_list(self._window_size)


class SimpleContext(Context):
    """A `SimpleContext` is a context in which each interaction is a separate and independent turn. The history of all previous turns is NOT saved.."""

    def add(self, c: Component | CBlock) -> SimpleContext:
        """Add a new component or CBlock to the context and return the updated context.

        Args:
            c (Component | CBlock): The component or content block to record.

        Returns:
            SimpleContext: A new ``SimpleContext`` containing only the added entry;
            prior history is not retained.
        """
        return SimpleContext.from_previous(self, c)

    def view_for_generation(self) -> list[Component | CBlock] | None:
        """Return an empty list, since ``SimpleContext`` does not pass history to the model.

        Each call to the model is treated as a stateless, independent exchange.
        No prior turns are forwarded.

        Returns:
            list[Component | CBlock] | None: Always an empty list.
        """
        return []



"""Hook payload classes for the Mellea plugin system."""

from .component import (
    ComponentPostErrorPayload,
    ComponentPostSuccessPayload,
    ComponentPreExecutePayload,
)
from .generation import GenerationPostCallPayload, GenerationPreCallPayload
from .sampling import (
    SamplingIterationPayload,
    SamplingLoopEndPayload,
    SamplingLoopStartPayload,
    SamplingRepairPayload,
)
from .session import (
    SessionCleanupPayload,
    SessionPostInitPayload,
    SessionPreInitPayload,
    SessionResetPayload,
)
from .tool import ToolPostInvokePayload, ToolPreInvokePayload
from .validation import ValidationPostCheckPayload, ValidationPreCheckPayload

__all__ = [
    # Component
    "ComponentPostErrorPayload",
    "ComponentPostSuccessPayload",
    "ComponentPreExecutePayload",
    # Generation
    "GenerationPostCallPayload",
    "GenerationPreCallPayload",
    # Sampling
    "SamplingIterationPayload",
    "SamplingLoopEndPayload",
    "SamplingLoopStartPayload",
    "SamplingRepairPayload",
    # Session
    "SessionCleanupPayload",
    "SessionPostInitPayload",
    "SessionPreInitPayload",
    "SessionResetPayload",
    # Tool
    "ToolPostInvokePayload",
    "ToolPreInvokePayload",
    # Validation
    "ValidationPostCheckPayload",
    "ValidationPreCheckPayload",
]



"""Core abstractions for the mellea library.

This package defines the fundamental interfaces and data structures on which every
other layer of mellea is built: the ``Backend``, ``Formatter``, and
``SamplingStrategy`` protocols; the ``Component``, ``CBlock``, ``Context``, and
``ModelOutputThunk`` data types that flow through the inference pipeline; and
``Requirement`` / ``ValidationResult`` for constrained generation. Start here when
building a new backend, formatter, or sampling strategy, or when you need the type
definitions shared across the library.
"""

from .backend import Backend, BaseModelSubclass, generate_walk
from .base import (
    C,
    CBlock,
    Component,
    ComponentParseError,
    Context,
    ContextTurn,
    GenerateLog,
    GenerateType,
    ImageBlock,
    ModelOutputThunk,
    ModelToolCall,
    S,
    TemplateRepresentation,
    blockify,
)
from .formatter import Formatter
from .requirement import Requirement, ValidationResult, default_output_to_bool
from .sampling import SamplingResult, SamplingStrategy
from .utils import FancyLogger

__all__ = [
    "Backend",
    "BaseModelSubclass",
    "C",
    "CBlock",
    "Component",
    "ComponentParseError",
    "Context",
    "ContextTurn",
    "FancyLogger",
    "Formatter",
    "GenerateLog",
    "GenerateType",
    "ImageBlock",
    "ModelOutputThunk",
    "ModelToolCall",
    "Requirement",
    "S",
    "SamplingResult",
    "SamplingStrategy",
    "TemplateRepresentation",
    "ValidationResult",
    "blockify",
    "default_output_to_bool",
    "generate_walk",
]



# pytest: ollama, llm

from mellea.core import Requirement
from mellea.stdlib.requirements import check, req, simple_validate

requirements: list[Requirement | str] = [
    req("The email should have a salutation"),  # == r1
    req(
        "Use only lower-case letters",
        validation_fn=simple_validate(lambda x: x.lower() == x),
    ),  # == r2
    check("Do not mention purple elephants."),  # == r3
]

import mellea
from mellea.stdlib.sampling import RejectionSamplingStrategy


def write_email(m: mellea.MelleaSession, name: str, notes: str) -> str:
    email_candidate = m.instruct(
        "Write an email to {{name}} using the notes following: {{notes}}.",
        requirements=requirements,
        user_variables={"name": name, "notes": notes},
        strategy=RejectionSamplingStrategy(loop_budget=5),
        return_sampling_results=True,
    )
    if email_candidate.success:
        return str(email_candidate.result)
    else:
        return email_candidate.sample_generations[0].value or ""


m = mellea.start_session()
print(
    write_email(
        m,
        "Olivia",
        "Olivia helped the lab over the last few weeks by organizing intern events, advertising the speaker series, and handling issues with snack delivery.",
    )
)



# SPDX-License-Identifier: Apache-2.0

"""Type definitions that are shared within the Granite 3 family of models."""

# Third Party
import pydantic
import pydantic_core

# First Party
from ..base.types import (
    AssistantMessage,
    ChatTemplateKwargs,
    Document,
    GraniteChatCompletion,
    NoDefaultsMixin,
    SystemMessage,
    UserMessage,
    VLLMExtraBody,
)


class Hallucination(pydantic.BaseModel):
    """Hallucination data as returned by the model output parser.

    Attributes:
        hallucination_id (str): Unique identifier for the hallucination entry.
        risk (str): Risk level of the hallucination, e.g. ``"low"`` or ``"high"``.
        reasoning (str | None): Optional model-provided reasoning for why this
            sentence was flagged.
        response_text (str): The portion of the response text that is flagged.
        response_begin (int): Start character offset of ``response_text`` within
            the full response string.
        response_end (int): End character offset (exclusive) of ``response_text``
            within the full response string.
    """

    hallucination_id: str
    risk: str
    reasoning: str | None = None
    response_text: str
    response_begin: int
    response_end: int


class Citation(pydantic.BaseModel):
    """Citation data as returned by the model output parser.

    Attributes:
        citation_id (str): Unique identifier assigned to this citation.
        doc_id (str): Identifier of the source document being cited.
        context_text (str): Verbatim text from the source document that is cited.
        context_begin (int): Start character offset of ``context_text`` within
            the source document.
        context_end (int): End character offset (exclusive) of ``context_text``
            within the source document.
        response_text (str): The portion of the response text that makes this
            citation.
        response_begin (int): Start character offset of ``response_text`` within
            the response string.
        response_end (int): End character offset (exclusive) of ``response_text``
            within the response string.
    """

    citation_id: str
    doc_id: str
    context_text: str
    context_begin: int
    context_end: int
    response_text: str
    response_begin: int
    response_end: int


class Granite3Controls(pydantic.BaseModel):
    """Control flags for Granite 3.x model output behaviour.

    Specifies which optional output features the model should produce, such as
    inline citations, hallucination risk annotations, response length constraints,
    and originality style.

    Attributes:
        citations (bool | None): When ``True``, instructs the model to annotate
            factual claims with inline citation markers.
        hallucinations (bool | None): When ``True``, instructs the model to
            append a list of sentences that may be hallucinated.
        length (str | None): Requested response length; must be ``"short"``,
            ``"long"``, or ``None`` for no constraint.
        originality (str | None): Requested response originality style; must be
            ``"extractive"``, ``"abstractive"``, or ``None``.
    """

    citations: bool | None = None
    hallucinations: bool | None = None
    length: str | None = None  # Length output control variable
    originality: str | None = None

    @pydantic.field_validator("length", mode="after")
    @classmethod
    def _validate_length(cls, value: str | None) -> str | None:
        if value is None or value == "short" or value == "long":
            return value
        raise pydantic_core.PydanticCustomError(
            "length field validator",
            'length ({length}) must be "short" or "long" or None',
            {"length": value},
        )

    @pydantic.field_validator("originality", mode="after")
    @classmethod
    def _validate_originality(cls, value: str | None) -> str | None:
        if value is None or value == "extractive" or value == "abstractive":
            return value
        raise pydantic_core.PydanticCustomError(
            "originality field validator",
            'originality ({originality}) must be "extractive" or "abstractive" or None',
            {"originality": value},
        )


class Granite3Kwargs(ChatTemplateKwargs, NoDefaultsMixin):
    """Chat template keyword arguments specific to IBM Granite 3.x models.

    Extends :class:`ChatTemplateKwargs` with Granite 3-specific options for
    output control flags and chain-of-thought (thinking) mode.

    Attributes:
        controls (Granite3Controls | None): Optional output control flags that
            enable or configure citations, hallucination detection, response
            length, and originality style.
        thinking (bool): When ``True``, enables chain-of-thought reasoning mode.
            Defaults to ``False``.
    """

    controls: Granite3Controls | None = None
    thinking: bool = False


class Granite3ChatCompletion(GraniteChatCompletion):
    """Class that represents the inputs common to IBM Granite 3.x models.

    Class that represents the inputs that are common to models of the IBM Granite 3.x
    family.
    """

    def controls(self) -> Granite3Controls:
        """Return the Granite 3 controls record for this chat completion request.

        Returns:
            Granite3Controls: The controls record from the chat template kwargs,
                or an empty :class:`Granite3Controls` if none were specified.
        """
        kwargs = self.extra_body.chat_template_kwargs if self.extra_body else None
        if kwargs and isinstance(kwargs, Granite3Kwargs) and kwargs.controls:
            return kwargs.controls
        return Granite3Controls()

    def thinking(self) -> bool:
        """Return whether chain-of-thought thinking mode is enabled.

        Returns:
            bool: ``True`` if the ``thinking`` flag is set in the chat template
                kwargs; ``False`` otherwise.
        """
        kwargs = self.extra_body.chat_template_kwargs if self.extra_body else None
        return bool(kwargs and isinstance(kwargs, Granite3Kwargs) and kwargs.thinking)

    @pydantic.field_validator("extra_body")
    @classmethod
    def _validate_chat_template_kwargs(cls, extra_body: VLLMExtraBody) -> VLLMExtraBody:
        """Validate Granite 3 chat template kwargs and convert to dataclass.

        Validates kwargs that are specific to Granite 3 chat templates and converts
        the ``chat_template_kwargs`` field to a Granite 3-specific dataclass.

        Other arguments are currently passed through without checking.
        """
        if extra_body.chat_template_kwargs:
            kwargs_dict = extra_body.chat_template_kwargs.model_dump()
            extra_body.chat_template_kwargs = Granite3Kwargs.model_validate(kwargs_dict)
        return extra_body

    @pydantic.field_validator("messages")
    @classmethod
    def _validate_inputs_messages(cls, messages: list) -> list:
        # Make a copy so the validation code below can mutate the messages list but pass
        # through the original value. The caller also might have a pointer to the list.
        original_messages = messages
        messages = messages.copy()

        # There is no supervised fine tuning data for the case of zero messages.
        # Models are not guaranteed to produce a valid response if there are zero
        # messages.
        if len(messages) == 0:
            raise ValueError(
                "No messages. Model behavior for this case is not defined."
            )

        # The first message, and only the first message, may be the system message.
        first_message_is_system_message = isinstance(messages[0], SystemMessage)
        if first_message_is_system_message:
            messages = messages[1:]
            # If there is a system message, there must be at least one more user or
            # assistant message.
            if len(messages) == 0:
                raise ValueError(
                    "Input contains only a system message. Model behavior for this "
                    "case is not defined."
                )

        # The first message that is not a system message must be
        # either a user or assistant message.
        if not isinstance(messages[0], UserMessage | AssistantMessage):
            if first_message_is_system_message:
                raise ValueError(
                    f"First message after system message must be a user or "
                    f"assistant message. Found type {type(messages[0])}"
                )
            raise ValueError(
                f"First message must be a system, user, or assistant "
                f"Found type {type(messages[0])}"
            )

        # Undocumented constraint: All other messages form a conversation that
        # alternates strictly between user and assistant, possibly with tool calls
        # after an assistant turn and before the next user turn.
        # TODO: Validate this invariant.

        # Pydantic will use the value that this validator returns as the value of the
        # messages field. Undo any changes that we made during validation and return
        # the original value.
        return original_messages


class Granite3AssistantMessage(AssistantMessage):
    """An assistant message with Granite 3 specific fields.

    Attributes:
        reasoning_content (str | None): Optional chain-of-thought reasoning text
            produced before the final response.
        citations (list[Citation] | None): Optional list of citations parsed from
            the model output.
        documents (list[Document] | None): Optional list of documents referenced
            in the model output.
        hallucinations (list[Hallucination] | None): Optional list of hallucination
            annotations parsed from the model output.
        stop_reason (str | None): Optional reason the model stopped generating.
    """

    reasoning_content: str | None = None
    citations: list[Citation] | None = None
    documents: list[Document] | None = None
    hallucinations: list[Hallucination] | None = None
    stop_reason: str | None = None

    raw_content: str | None = pydantic.Field(
        default=None,
        description=(
            "Raw response content without any parsing, for debugging and "
            "re-serialization."
        ),
    )



"""Abstract ``Backend`` interface and generation-walk utilities.

Defines the ``Backend`` abstract base class whose two key abstract methods —
``generate_from_context`` (context-aware single-action generation) and
``generate_from_raw`` (context-free batch generation) — all concrete backends must
implement. Also provides ``generate_walk``, which traverses a ``Component`` tree to
find un-computed ``ModelOutputThunk`` leaves that need to be resolved before rendering.
"""

import abc
import asyncio
import functools
import itertools
import time
from collections.abc import Sequence
from typing import final, overload

import pydantic
import typing_extensions

from ..plugins.manager import has_plugins, invoke_hook
from ..plugins.types import HookType
from .base import C, CBlock, Component, Context, ModelOutputThunk
from .utils import FancyLogger

# Necessary to define a type variable that has a default value.
# This is because VSCode's pyright static type checker instantiates
# all type parameters, order them in a specific order, and checks
# that a type parameter with a default is not followed by a type parameter without a default.
# This was originally specified in PEP 696 (Type Defaults for Type Parameters) and
# is now part of official spec https://typing.python.org/en/latest/spec/generics.html#type-parameter-defaults .
# For example, in mellea/stdlib/functional.py ,
#    def act(action: Component[S], ... format: type[BaseModelSubclass] | None = None, ...)
# gets instantiated as act[S, BaseModelSubclass].
# S is TypeVar("S", default=Any, covariant=True) which has a default.
#
BaseModelSubclass = typing_extensions.TypeVar(
    "BaseModelSubclass", bound=pydantic.BaseModel, default=pydantic.BaseModel
)  # must be a subclass of BaseModel


class Backend(abc.ABC):
    """Abstract base class for all inference backends.

    All concrete backends must implement ``generate_from_context`` (context-aware
    single-action generation) and ``generate_from_raw`` (context-free batch
    generation). The ``do_generate_walk`` / ``do_generate_walks`` helpers can be
    used to pre-compute any unresolved ``ModelOutputThunk`` leaves before rendering.
    """

    @final
    async def generate_from_context(
        self,
        action: Component[C] | CBlock,
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> tuple[ModelOutputThunk[C], Context]:
        """Generates a model output from a context. May not mutate the context. This must be called from a running event loop as it creates a task to run the generation request.

        Args:
            action: The last item of the context should be passed in as an `action` instead of as part of the `ctx`. See `docs/dev/generate_signature_decisions.md`.
            ctx: The rest of the context.
            format: A response format to used for structured outputs / constrained decoding.
            model_options: Any model options to upsert into the defaults for this call.
            tool_calls: If `True`, then tool calls are extracts from the `action` `Component`. Assumption: if tool_calls is enabled, then the action `Component` has a TemplateRepresentation

        Returns:
            a tuple of (ModelOutputThunk, Context) where the Context is the new context after the generation has been completed.
        """
        # --- generation_pre_call hook ---
        if has_plugins(HookType.GENERATION_PRE_CALL):
            from ..plugins.hooks.generation import GenerationPreCallPayload

            pre_payload = GenerationPreCallPayload(
                action=action,
                context=ctx,
                format=format,
                model_options=model_options or {},
                tool_calls=tool_calls,
            )
            _, pre_payload = await invoke_hook(
                HookType.GENERATION_PRE_CALL, pre_payload, backend=self
            )
            model_options = pre_payload.model_options
            format = pre_payload.format
            tool_calls = pre_payload.tool_calls

        return await self._generate_from_context(
            action,
            ctx,
            format=format,
            model_options=model_options,
            tool_calls=tool_calls,
        )

    @abc.abstractmethod
    async def _generate_from_context(
        self,
        action: Component[C] | CBlock,
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> tuple[ModelOutputThunk[C], Context]:
        """Backend implementers should override this method to generate the actual response.

        Args:
            action: The last item of the context should be passed in as an `action` instead of as part of the `ctx`. See `docs/dev/generate_signature_decisions.md`.
            ctx: The rest of the context.
            format: A response format to used for structured outputs / constrained decoding.
            model_options: Any model options to upsert into the defaults for this call.
            tool_calls: If `True`, then tool calls are extracts from the `action` `Component`. Assumption: if tool_calls is enabled, then the action `Component` has a TemplateRepresentation

        Returns:
            a tuple of (ModelOutputThunk, Context) where the Context is the new context after the generation has been completed.
        """
        ...

    @overload
    async def generate_from_raw(
        self,
        actions: list[Component[C]],
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> list[ModelOutputThunk[C]]: ...

    @overload
    async def generate_from_raw(
        self,
        actions: list[Component[C] | CBlock],
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> list[ModelOutputThunk[C | str]]: ...

    @abc.abstractmethod
    async def generate_from_raw(
        self,
        actions: Sequence[Component[C] | CBlock],
        ctx: Context,
        *,
        format: type[BaseModelSubclass] | None = None,
        model_options: dict | None = None,
        tool_calls: bool = False,
    ) -> list[ModelOutputThunk]:
        """Generates a model output from the provided input. Does not use context or templates.

        Args:
            actions: list of actions to generate responses for. Each action is separate.
            ctx: context passed to generation. Currently not used in generate_from_raw
            format: A response format to used for structured outputs / constrained decoding. Note: some backends do not support this parameter. They will log warnings and continue to generate.
            model_options: Any model options to upsert into the defaults for this call.
            tool_calls: Always set to false unless supported by backend.

        Returns:
            list[ModelOutputThunk]: A list of output thunks, one per action, in the same order as ``actions``.
        """

    async def do_generate_walk(
        self, action: CBlock | Component | ModelOutputThunk
    ) -> None:
        """Awaits all uncomputed ``ModelOutputThunk`` leaves reachable from ``action``.

        Traverses the component tree rooted at ``action`` via ``generate_walk``, collects
        any uncomputed ``ModelOutputThunk`` nodes, and concurrently awaits them all.

        Args:
            action (CBlock | Component | ModelOutputThunk): The root node to traverse.
        """
        _to_compute = list(generate_walk(action))
        coroutines = [x.avalue() for x in _to_compute]
        # The following log message might get noisy. Feel free to remove if so.
        if len(_to_compute) > 0:
            FancyLogger.get_logger().info(
                f"generate_from_chat_context awaited on {len(_to_compute)} uncomputed mots."
            )
        await asyncio.gather(*coroutines)

    async def do_generate_walks(
        self, actions: list[CBlock | Component | ModelOutputThunk]
    ) -> None:
        """Awaits all uncomputed ``ModelOutputThunk`` leaves reachable from each action in ``actions``.

        Traverses the component tree of every action in the list via ``generate_walk``, collects
        all uncomputed ``ModelOutputThunk`` nodes across all actions, and concurrently awaits them.

        Args:
            actions (list[CBlock | Component | ModelOutputThunk]): The list of root nodes to traverse.
        """
        _to_compute = []
        for action in actions:
            _to_compute.extend(list(generate_walk(action)))
        coroutines = [x.avalue() for x in _to_compute]
        # The following log message might get noisy. Feel free to remove if so.
        if len(_to_compute) > 0:
            FancyLogger.get_logger().info(
                f"generate_from_chat_context awaited on {len(_to_compute)} uncomputed mots."
            )
        await asyncio.gather(*coroutines)


def generate_walk(c: CBlock | Component | ModelOutputThunk) -> list[ModelOutputThunk]:
    """Return all uncomputed ``ModelOutputThunk`` leaves reachable from ``c``.

    Args:
        c: A ``CBlock``, ``Component``, or ``ModelOutputThunk`` to traverse.

    Returns:
        A flat list of uncomputed ``ModelOutputThunk`` instances in the order
        they need to be resolved (depth-first over ``Component.parts()``).

    Raises:
        ValueError: If any element encountered during traversal is not a ``CBlock``,
            ``Component``, or ``ModelOutputThunk``.
    """
    match c:
        case ModelOutputThunk() if not c.is_computed():
            return [c]
        case CBlock():
            return []
        case Component():
            parts_walk = [generate_walk(p) for p in c.parts()]
            return list(itertools.chain.from_iterable(parts_walk))  # aka flatten
        case _:
            raise ValueError(
                f"parts should only contain CBlocks, Components, or ModelOutputThunks; found `{c!s:.10}{'...' if len(str(c)) > 10 else ''}` (type: {type(c)})"
            )



# pytest: skip
# SKIP REASON: Requires an AWS bearer token for Bedrock.
#!/usr/bin/env -S uv run --script
# /// script
# requires-python = ">=3.11"
# dependencies = [
#   "mellea[litellm]",
#   "boto3" # including so that this example works before the next release.
# ]
# ///
import os

import mellea

try:
    import boto3
except Exception:
    raise Exception(
        "Using Bedrock requires separately installing boto3. "
        "Run `uv pip install mellea[litellm]`"
    )

assert "AWS_BEARER_TOKEN_BEDROCK" in os.environ.keys(), (
    "Using AWS Bedrock requires setting a AWS_BEARER_TOKEN_BEDROCK environment variable. "
    "Generate a key from the AWS console at: https://us-east-1.console.aws.amazon.com/bedrock/home?region=us-east-1#/api-keys?tab=long-term "
    "Then run `export AWS_BEARER_TOKEN_BEDROCK=<insert your key here>"
)

MODEL_ID = "bedrock/converse/us.amazon.nova-pro-v1:0"

m = mellea.start_session(backend_name="litellm", model_id=MODEL_ID)

result = m.chat("Give me three facts about Amazon.")

print(result.content)



"""Logging utilities for the mellea core library.

Provides ``FancyLogger``, a singleton logger with colour-coded console output and
an optional REST handler (``RESTHandler``) that forwards log records to a local
``/api/receive`` endpoint when the ``FLOG`` environment variable is set. All
internal mellea modules obtain their logger via ``FancyLogger.get_logger()``.
"""

import json
import logging
import os
import sys

import requests


class RESTHandler(logging.Handler):
    """Logging handler that forwards records to a local REST endpoint.

    Sends log records as JSON to ``/api/receive`` when the ``FLOG`` environment
    variable is set. Failures are silently suppressed to avoid disrupting the
    application.

    Args:
        api_url (str): The URL of the REST endpoint that receives log records.
        method (str): HTTP method to use when sending records (default ``"POST"``).
        headers (dict | None): HTTP headers to send; defaults to
            ``{"Content-Type": "application/json"}`` when ``None``.
    """

    def __init__(
        self, api_url: str, method: str = "POST", headers: dict[str, str] | None = None
    ) -> None:
        """Initializes a RESTHandler; uses application/json by default."""
        super().__init__()
        self.api_url = api_url
        self.method = method
        self.headers = headers or {"Content-Type": "application/json"}

    def emit(self, record: logging.LogRecord) -> None:
        """Forwards a log record to the REST endpoint when the ``FLOG`` environment variable is set.

        Silently suppresses any network or HTTP errors to avoid disrupting the application.

        Args:
            record (logging.LogRecord): The log record to forward.
        """
        if os.environ.get("FLOG"):
            log_data = self.format(record)
            try:
                response = requests.request(
                    self.method,
                    self.api_url,
                    headers=self.headers,
                    # data=json.dumps([{"log": log_data}]),
                    data=json.dumps([log_data]),
                )
                response.raise_for_status()
            except requests.exceptions.RequestException as _:
                pass


class JsonFormatter(logging.Formatter):
    """Logging formatter that serialises log records as structured JSON dicts.

    Includes timestamp, level, message, module, function name, line number,
    process ID, thread ID, and (if present) exception information.
    """

    def format(self, record):  # type: ignore
        """Formats a log record as a JSON-serialisable dictionary.

        Includes timestamp, level, message, module, function name, line number,
        process ID, thread ID, and exception info if present.

        Args:
            record (logging.LogRecord): The log record to format.
        """
        log_record = {
            "timestamp": self.formatTime(record, self.datefmt),
            "level": record.levelname,
            "message": record.getMessage(),
            "module": record.module,
            "function": record.funcName,
            "line_number": record.lineno,
            "process_id": record.process,
            "thread_id": record.thread,
        }
        if record.exc_info:
            log_record["exception"] = self.formatException(record.exc_info)
        return log_record


class CustomFormatter(logging.Formatter):
    """A nice custom formatter copied from [Sergey Pleshakov's post on StackOverflow](https://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output).

    Attributes:
        cyan (str): ANSI escape code for cyan text, used for DEBUG messages.
        grey (str): ANSI escape code for grey text, used for INFO messages.
        yellow (str): ANSI escape code for yellow text, used for WARNING messages.
        red (str): ANSI escape code for red text, used for ERROR messages.
        bold_red (str): ANSI escape code for bold red text, used for CRITICAL messages.
        reset (str): ANSI escape code to reset text colour.
        FORMATS (dict): Mapping from logging level integer to the colour-formatted format string.
    """

    cyan = "\033[96m"  # Cyan
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    _format_string = "=== %(asctime)s-%(levelname)s ======\n%(message)s"

    FORMATS = {
        logging.DEBUG: cyan + _format_string + reset,
        logging.INFO: grey + _format_string + reset,
        logging.WARNING: yellow + _format_string + reset,
        logging.ERROR: red + _format_string + reset,
        logging.CRITICAL: bold_red + _format_string + reset,
    }

    def format(self, record: logging.LogRecord) -> str:
        """Formats a log record using a colour-coded ANSI format string based on the record's log level.

        Args:
            record (logging.LogRecord): The log record to format.

        Returns:
            str: The formatted log record string with ANSI colour codes applied.
        """
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, datefmt="%H:%M:%S")
        return formatter.format(record)


class FancyLogger:
    """Singleton logger with colour-coded console output and optional REST forwarding.

    Obtain the shared logger instance via ``FancyLogger.get_logger()``. Log level
    defaults to ``INFO`` but can be raised to ``DEBUG`` by setting the ``DEBUG``
    environment variable. When the ``FLOG`` environment variable is set, records are
    also forwarded to a local ``/api/receive`` REST endpoint via ``RESTHandler``.

    Attributes:
        logger (logging.Logger | None): The shared ``logging.Logger`` instance; ``None`` until first call to ``get_logger()``.
        CRITICAL (int): Numeric level for critical log messages (50).
        FATAL (int): Alias for ``CRITICAL`` (50).
        ERROR (int): Numeric level for error log messages (40).
        WARNING (int): Numeric level for warning log messages (30).
        WARN (int): Alias for ``WARNING`` (30).
        INFO (int): Numeric level for informational log messages (20).
        DEBUG (int): Numeric level for debug log messages (10).
        NOTSET (int): Numeric level meaning no level is set (0).
    """

    logger = None

    CRITICAL = 50
    FATAL = CRITICAL
    ERROR = 40
    WARNING = 30
    WARN = WARNING
    INFO = 20
    DEBUG = 10
    NOTSET = 0

    @staticmethod
    def get_logger() -> logging.Logger:
        """Returns a FancyLogger.logger and sets level based upon env vars.

        Returns:
            Configured logger with REST, stream, and optional OTLP handlers.
        """
        if FancyLogger.logger is None:
            logger = logging.getLogger("fancy_logger")
            # Only set default level if user hasn't already configured it
            if logger.level == logging.NOTSET:
                if os.environ.get("DEBUG"):
                    logger.setLevel(FancyLogger.DEBUG)
                else:
                    logger.setLevel(FancyLogger.INFO)

            # Define REST API endpoint
            api_url = "http://localhost:8000/api/receive"

            # Create REST handler
            rest_handler = RESTHandler(api_url)

            # Create formatter and set it for the handler
            # formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
            rest_handler.setFormatter(JsonFormatter())

            # Add handler to the logger
            logger.addHandler(rest_handler)

            stream_handler = logging.StreamHandler(stream=sys.stdout)
            # stream_handler.setLevel(logging.INFO)
            stream_handler.setFormatter(CustomFormatter(datefmt="%H:%M:%S,%03d"))
            logger.addHandler(stream_handler)

            # Add OTLP handler if enabled
            from ..telemetry import get_otlp_log_handler

            otlp_handler = get_otlp_log_handler()
            if otlp_handler:
                otlp_handler.setFormatter(JsonFormatter())
                logger.addHandler(otlp_handler)

            FancyLogger.logger = logger
        return FancyLogger.logger



# pytest: skip_always

import mellea
from mellea.stdlib.components.docs.richdocument import RichDocument

m = mellea.start_session()

doc = RichDocument.from_document_file(
    "https://www.fs.usda.gov/foresthealth/docs/fidls/FIDL-78-ArmillariaRootDisease.pdf"
)

print(
    m.instruct(
        "Determine whether the mellea fungus only feeds on dead or weakened trees, thereby improving forest health",
        grounding_context={"usfs": doc},
    ).value
)



"""OpenTelemetry tracing instrumentation for Mellea.

This module provides two independent trace scopes:
1. Application Trace (mellea.application) - User-facing operations
2. Backend Trace (mellea.backend) - LLM backend interactions

Follows OpenTelemetry Gen-AI semantic conventions:
https://opentelemetry.io/docs/specs/semconv/gen-ai/

Configuration via environment variables:
- MELLEA_TRACE_APPLICATION: Enable/disable application tracing (default: false)
- MELLEA_TRACE_BACKEND: Enable/disable backend tracing (default: false)
- OTEL_EXPORTER_OTLP_ENDPOINT: OTLP endpoint for trace export
- OTEL_SERVICE_NAME: Service name for traces (default: mellea)
"""

import os
from collections.abc import Generator
from contextlib import contextmanager
from importlib.metadata import version
from typing import Any

# Try to import OpenTelemetry, but make it optional
try:
    from opentelemetry import trace
    from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
    from opentelemetry.sdk.resources import Resource
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter

    _OTEL_AVAILABLE = True
except ImportError:
    _OTEL_AVAILABLE = False
    # Provide dummy types for type hints
    trace = None  # type: ignore

# Configuration from environment variables
# Disable tracing if OpenTelemetry is not available
_TRACE_APPLICATION_ENABLED = _OTEL_AVAILABLE and os.getenv(
    "MELLEA_TRACE_APPLICATION", "false"
).lower() in ("true", "1", "yes")
_TRACE_BACKEND_ENABLED = _OTEL_AVAILABLE and os.getenv(
    "MELLEA_TRACE_BACKEND", "false"
).lower() in ("true", "1", "yes")
_OTLP_ENDPOINT = os.getenv("OTEL_EXPORTER_OTLP_ENDPOINT")
_SERVICE_NAME = os.getenv("OTEL_SERVICE_NAME", "mellea")
_CONSOLE_EXPORT = os.getenv("MELLEA_TRACE_CONSOLE", "false").lower() in (
    "true",
    "1",
    "yes",
)


def _setup_tracer_provider() -> Any:
    """Set up the global tracer provider with OTLP exporter if configured."""
    if not _OTEL_AVAILABLE:
        return None

    resource = Resource.create({"service.name": _SERVICE_NAME})  # type: ignore
    provider = TracerProvider(resource=resource)  # type: ignore

    # Add OTLP exporter if endpoint is configured
    if _OTLP_ENDPOINT:
        otlp_exporter = OTLPSpanExporter(endpoint=_OTLP_ENDPOINT)  # type: ignore
        provider.add_span_processor(BatchSpanProcessor(otlp_exporter))  # type: ignore

    # Add console exporter for debugging if enabled
    # Note: Console exporter may cause harmless errors during test cleanup
    if _CONSOLE_EXPORT:
        try:
            console_exporter = ConsoleSpanExporter()  # type: ignore
            provider.add_span_processor(BatchSpanProcessor(console_exporter))  # type: ignore
        except Exception:
            # Silently ignore console exporter setup failures
            pass

    trace.set_tracer_provider(provider)  # type: ignore
    return provider


# Initialize tracer provider if any tracing is enabled
_tracer_provider = None
_application_tracer = None
_backend_tracer = None

if _OTEL_AVAILABLE and (_TRACE_APPLICATION_ENABLED or _TRACE_BACKEND_ENABLED):
    _tracer_provider = _setup_tracer_provider()
    # Create separate tracers for application and backend
    _mellea_version = version("mellea")
    _application_tracer = trace.get_tracer("mellea.application", _mellea_version)  # type: ignore
    _backend_tracer = trace.get_tracer("mellea.backend", _mellea_version)  # type: ignore


def is_application_tracing_enabled() -> bool:
    """Check if application tracing is enabled.

    Returns:
        True if application tracing has been enabled via the
        ``MELLEA_TRACE_APPLICATION`` environment variable.
    """
    return _TRACE_APPLICATION_ENABLED


def is_backend_tracing_enabled() -> bool:
    """Check if backend tracing is enabled.

    Returns:
        True if backend tracing has been enabled via the
        ``MELLEA_TRACE_BACKEND`` environment variable.
    """
    return _TRACE_BACKEND_ENABLED


@contextmanager
def trace_application(name: str, **attributes: Any) -> Generator[Any, None, None]:
    """Create an application trace span if application tracing is enabled.

    Args:
        name: Name of the span.
        **attributes: Additional attributes to add to the span.

    Yields:
        The span object if tracing is enabled, otherwise ``None``.
    """
    if _TRACE_APPLICATION_ENABLED and _application_tracer is not None:
        with _application_tracer.start_as_current_span(name) as span:  # type: ignore
            for key, value in attributes.items():
                if value is not None:
                    _set_attribute_safe(span, key, value)
            yield span
    else:
        yield None


@contextmanager
def trace_backend(name: str, **attributes: Any) -> Generator[Any, None, None]:
    """Create a backend trace span if backend tracing is enabled.

    Follows Gen-AI semantic conventions for LLM operations.

    Args:
        name: Name of the span.
        **attributes: Additional attributes to add to the span.

    Yields:
        The span object if tracing is enabled, otherwise ``None``.
    """
    if _TRACE_BACKEND_ENABLED and _backend_tracer is not None:
        with _backend_tracer.start_as_current_span(name) as span:  # type: ignore
            # Set Gen-AI operation type
            span.set_attribute("gen_ai.operation.name", name)

            for key, value in attributes.items():
                if value is not None:
                    _set_attribute_safe(span, key, value)
            yield span
    else:
        yield None


def start_backend_span(name: str, **attributes: Any) -> Any:
    """Start a backend trace span without auto-closing (for async operations).

    Use this when you need to manually control span lifecycle, such as for
    async operations where the span should remain open until post-processing.

    Args:
        name: Name of the span
        **attributes: Additional attributes to add to the span

    Returns:
        The span object if tracing is enabled, otherwise None
    """
    if _TRACE_BACKEND_ENABLED and _backend_tracer is not None:
        span = _backend_tracer.start_span(name)  # type: ignore
        # Set Gen-AI operation type
        span.set_attribute("gen_ai.operation.name", name)

        for key, value in attributes.items():
            if value is not None:
                _set_attribute_safe(span, key, value)
        return span
    return None


def end_backend_span(span: Any) -> None:
    """End a backend trace span.

    Args:
        span: The span object to end
    """
    if span is not None:
        span.end()


def _set_attribute_safe(span: Any, key: str, value: Any) -> None:
    """Set an attribute on a span, handling type conversions.

    Args:
        span: The span object
        key: Attribute key
        value: Attribute value (will be converted to appropriate type)
    """
    if value is None:
        return

    # Handle different value types according to OpenTelemetry spec
    if isinstance(value, bool):
        span.set_attribute(key, value)
    elif isinstance(value, int | float):
        span.set_attribute(key, value)
    elif isinstance(value, str):
        span.set_attribute(key, value)
    elif isinstance(value, list | tuple):
        # Convert to list of strings
        span.set_attribute(key, [str(v) for v in value])
    else:
        # Convert other types to string
        span.set_attribute(key, str(value))


def set_span_attribute(span: Any, key: str, value: Any) -> None:
    """Set an attribute on a span if the span is not None.

    Args:
        span: The span object (may be None if tracing is disabled)
        key: Attribute key
        value: Attribute value
    """
    if span is not None and value is not None:
        _set_attribute_safe(span, key, value)


def set_span_error(span: Any, exception: Exception) -> None:
    """Record an exception on a span if the span is not None.

    Args:
        span: The span object (may be None if tracing is disabled)
        exception: The exception to record
    """
    if span is not None and _OTEL_AVAILABLE:
        span.record_exception(exception)
        span.set_status(trace.Status(trace.StatusCode.ERROR, str(exception)))  # type: ignore


__all__ = [
    "end_backend_span",
    "is_application_tracing_enabled",
    "is_backend_tracing_enabled",
    "set_span_attribute",
    "set_span_error",
    "start_backend_span",
    "trace_application",
    "trace_backend",
]



import pytest

from mellea.core import CBlock, ModelOutputThunk, default_output_to_bool
from mellea.stdlib.context import ChatContext, Context
from mellea.stdlib.requirements import (
    as_markdown_list,
    is_markdown_list,
    is_markdown_table,
)


def from_model(s: str) -> Context:
    ctx = ChatContext()
    ctx = ctx.add(ModelOutputThunk(value=s, meta={"test": True}))
    return ctx


MARKDOWN_LIST_CTX: Context = from_model(
    """
 * This is the first item
 * This is the second item
 * This is the third item
"""
)

MARKDOWN_OL_LIST_CTX: Context = from_model(
    """
1. The first item is the most important.
2. The second item is not as important.
3. Silly things get buried in the middle.
4. But the last item we save for being also important.
"""
)


MARKDOWN_TABLE_CONTEXT: Context = from_model(
    """
| Month    | Revenue |
| -------- | ------- |
| January  | $100    |
| February | $900    |
| March    | $14     |
"""
)


async def test_markdown_list():
    result = await is_markdown_list.validate(None, MARKDOWN_LIST_CTX)  # type: ignore
    assert result
    assert len(as_markdown_list(MARKDOWN_LIST_CTX)) == 3  # type: ignore
    assert len(as_markdown_list(MARKDOWN_OL_LIST_CTX)) == 4  # type: ignore
    assert type(as_markdown_list(MARKDOWN_OL_LIST_CTX)[0]) is str  # type: ignore
    result = await is_markdown_list.validate(None, MARKDOWN_OL_LIST_CTX)  # type: ignore
    assert result


async def test_markdown_table():
    result = await is_markdown_table.validate(None, MARKDOWN_TABLE_CONTEXT)  # type: ignore
    assert result


def test_default_output_to_bool_yes():
    assert default_output_to_bool("yeS")


def test_default_output_to_bool_no():
    assert not default_output_to_bool("nO")


def test_default_output_to_bool_complicated_yes():
    assert default_output_to_bool(
        CBlock("The requirement is met by the output. Therefore, my answer is yes.")
    )


def test_default_output_to_bool_word_with_yes_in_it():
    assert not default_output_to_bool(
        "Here's a word that meets those requirements: ayes."
    )


if __name__ == "__main__":
    pytest.main([__file__])



import asyncio
import json
from typing import Annotated

import pydantic
import pytest

from mellea import start_session
from mellea.backends import ModelOption
from mellea.backends.ollama import OllamaModelBackend
from mellea.core import CBlock, Requirement
from mellea.stdlib.context import SimpleContext
from mellea.stdlib.requirements import simple_validate

# Mark all tests in this module as requiring Ollama
pytestmark = [pytest.mark.ollama, pytest.mark.llm]


@pytest.fixture(scope="function")
def session():
    """Fresh Ollama session for each test."""
    session = start_session()
    yield session
    session.reset()


@pytest.mark.qualitative
def test_simple_instruct(session) -> None:
    result = session.instruct(
        "Write an email to Hendrik trying to sell him self-sealing stembolts."
    )
    assert result.value.startswith("Subject")
    assert "chat_response" in result._meta
    assert result._meta["chat_response"].message.role == "assistant"

    assert isinstance(result.parsed_repr, str)


@pytest.mark.qualitative
def test_instruct_with_requirement(session) -> None:
    session.instruct(
        "Write an email to Hendrik convincing him to buy some self-sealing stembolts."
    )

    email_word_count_req = Requirement(
        "The email should be at most 100",
        validation_fn=simple_validate(lambda x: len(" ".split(x)) <= 100),
    )

    happy_tone_req = Requirement(
        "The email should sound happy in tone.",
        output_to_bool=lambda x: "happy" in x.value,  # type: ignore
    )

    sad_tone_req = Requirement("The email should sound sad in tone.")

    results = session.validate(
        reqs=[email_word_count_req, happy_tone_req, sad_tone_req]
    )
    print(results)


@pytest.mark.qualitative
def test_chat(session) -> None:
    output_message = session.chat("What is 1+1?")
    assert "2" in output_message.content, (
        f"Expected a message with content containing 2 but found {output_message}"
    )


@pytest.mark.qualitative
def test_format(session) -> None:
    class Person(pydantic.BaseModel):
        name: str
        # it does not support regex patterns in json schema
        email_address: str
        # email_address: Annotated[
        #     str,
        #     pydantic.StringConstraints(pattern=r"[a-zA-Z]{5,10}@example\.com"),
        # ]

    class Email(pydantic.BaseModel):
        to: Person
        subject: str
        body: str

    output = session.instruct(
        "Write a short email to Olivia, thanking her for organizing a sailing activity. Her email server is example.com. No more than two sentences. ",
        format=Email,
        model_options={ModelOption.MAX_NEW_TOKENS: 2**10},
    )
    print("Formatted output:")
    email = Email.model_validate_json(
        output.value
    )  # this should succeed because the output should be JSON because we passed in a format= argument...
    print(email)

    print("address:", email.to.email_address)
    # this is not guaranteed, due to the lack of regexp pattern
    # assert "@" in email.to.email_address
    # assert email.to.email_address.endswith("example.com")


@pytest.mark.qualitative
@pytest.mark.timeout(150)
async def test_generate_from_raw(session) -> None:
    # Note capital letter "W" at the beginning of each prompt. This capital letter is
    # very important to the ollama version of Granite 4.0 micro, the current default
    # model for Mellea.
    prompts = ["What is 1+1?", "What is 2+2?", "What is 3+3?", "What is 4+4?"]

    results = await session.backend.generate_from_raw(
        actions=[CBlock(value=prompt) for prompt in prompts],
        ctx=session.ctx,
        model_options={
            ModelOption.CONTEXT_WINDOW: 2048,
            # With raw prompts and high temperature, a response of arbitrary
            # length is normal operation.
            ModelOption.MAX_NEW_TOKENS: 100,
        },
    )

    assert len(results) == len(prompts)
    assert all(r.value for r in results), (
        f"One or more requests returned empty (possible backend timeout): {[r.value for r in results]}"
    )


@pytest.mark.xfail(reason="ollama sometimes fails generated structured outputs")
async def test_generate_from_raw_with_format(session) -> None:
    prompts = ["what is 1+1?", "what is 2+2?", "what is 3+3?", "what is 4+4?"]

    class Answer(pydantic.BaseModel):
        name: str
        value: int

    results = await session.backend.generate_from_raw(
        actions=[CBlock(value=prompt) for prompt in prompts],
        ctx=session.ctx,
        format=Answer,
        model_options={ModelOption.CONTEXT_WINDOW: 2048},
    )

    assert len(results) == len(prompts)
    assert all(r.value for r in results), (
        f"One or more requests returned empty (possible backend timeout): {[r.value for r in results]}"
    )

    for result in results:
        try:
            Answer.model_validate_json(result.value)
        except pydantic.ValidationError as e:
            assert False, f"formatting directive failed for {result.value}: {e.json()}"


async def test_async_parallel_requests(session) -> None:
    model_opts = {ModelOption.STREAM: True}
    mot1, _ = await session.backend.generate_from_context(
        CBlock("Say Hello."), SimpleContext(), model_options=model_opts
    )
    mot2, _ = await session.backend.generate_from_context(
        CBlock("Say Goodbye!"), SimpleContext(), model_options=model_opts
    )

    m1_val = None
    m2_val = None
    if not mot1.is_computed():
        m1_val = await mot1.astream()
    if not mot2.is_computed():
        m2_val = await mot2.astream()

    assert m1_val is not None, "should be a string val after generation"
    assert m2_val is not None, "should be a string val after generation"

    m1_final_val = await mot1.avalue()
    m2_final_val = await mot2.avalue()

    # Ideally, we would be able to assert that m1_final_val != m1_val, but sometimes the first streaming response
    # contains the full response.
    assert m1_final_val.startswith(m1_val), (
        "final val should contain the first streamed chunk"
    )
    assert m2_final_val.startswith(m2_val), (
        "final val should contain the first streamed chunk"
    )

    assert m1_final_val == mot1.value
    assert m2_final_val == mot2.value


async def test_async_avalue(session) -> None:
    mot1, _ = await session.backend.generate_from_context(
        CBlock("Say Hello."), SimpleContext()
    )
    m1_final_val = await mot1.avalue()
    assert m1_final_val is not None
    assert m1_final_val == mot1.value

    # Verify telemetry fields are populated
    assert mot1.usage is not None
    assert mot1.usage["prompt_tokens"] >= 0
    assert mot1.usage["completion_tokens"] > 0
    assert mot1.usage["total_tokens"] > 0
    assert isinstance(mot1.model, str)
    assert mot1.provider == "ollama"


def test_multiple_asyncio_runs(session) -> None:
    async def test():
        result = await session.achat("hello")
        assert result is not None

    asyncio.run(test())
    asyncio.run(test())


def test_client_cache(session) -> None:
    backend: OllamaModelBackend = session.backend
    first_client = backend._async_client

    async def get_client_async():
        return backend._async_client

    second_client = asyncio.run(get_client_async())

    items_in_cache = backend._client_cache.cache.values()
    assert len(items_in_cache) == 2, (
        "should be two clients in the cache since _async_client was called from two event loops"
    )
    assert first_client in items_in_cache
    assert second_client in items_in_cache

    third_client = backend._async_client
    assert third_client == first_client, (
        "clients in sync code should be the same if haven't been pushed out of the cache"
    )

    fourth_client = asyncio.run(get_client_async())
    assert fourth_client in backend._client_cache.cache.values()
    assert len(backend._client_cache.cache.values()) == 2


if __name__ == "__main__":
    pytest.main([__file__])



#!/usr/bin/env python3
"""Validate Mellea documentation: links and Python code snippets.

Standalone script — no dependencies beyond Python 3.10+ stdlib.
Idempotent: read-only, reports problems to stdout, exits non-zero
if any hard errors are found.

Usage
-----
    python docs/scripts/check_docs.py                # run all checks
    python docs/scripts/check_docs.py links           # links only
    python docs/scripts/check_docs.py code            # code only
    python docs/scripts/check_docs.py shell           # shell quoting only
    python docs/scripts/check_docs.py --verbose       # show every item checked

Link checks
-----------
* Internal doc-to-doc links (relative paths within docs/docs/).
* Mintlify absolute paths (/getting-started/installation etc.) resolved
  against docs/docs/ and docs.json navigation.
* Mintlify Card href="..." attributes (JSX).
* Links that escape docs/docs/ (e.g. ../../examples/) — these resolve
  on the local filesystem but NOT on the published Mintlify site.  They
  are flagged as errors: use a full GitHub URL instead.
* External URLs (https://) — checked with a lightweight HEAD request.
  Failures are reported as warnings (network-dependent).
* docs.json navbar links and nav page slugs.

Code checks
-----------
* Syntax — every ```python block is compiled with compile().
  Snippets that fail only because of `await` outside a function or
  leading indentation are classified as *fragments* (warning, not error).
* Import analysis — top-level imports are checked for availability.
  mellea.* imports are checked against the repo source tree.
  Third-party imports that can't be found produce warnings.
* Missing-import heuristic — flags known mellea names used but never
  imported.
* Duplicate detection — code blocks of 4+ non-blank lines that appear
  identically in different files are flagged for consolidation.

Shell checks
------------
* Scans ```bash / ```shell blocks for `pip install X[extras]` or
  `uv pip install X[extras]` without shell quoting.  Unquoted square
  brackets break in zsh.
"""

from __future__ import annotations

import argparse
import ast
import hashlib
import importlib.util
import json
import re
import ssl
import sys
import urllib.error
import urllib.request
from pathlib import Path

# ---------------------------------------------------------------------------
# Configuration
# ---------------------------------------------------------------------------

SCRIPT_DIR = Path(__file__).resolve().parent
REPO_ROOT = SCRIPT_DIR.parent.parent  # docs/scripts/../../
DOCS_ROOT = REPO_ROOT / "docs" / "docs"  # Mintlify content root

# Skip API reference pages (separate PR)
SKIP_PREFIXES = ("api/",)

# GitHub base for converting escaped relative links
GITHUB_BASE = "https://github.com/generative-computing/mellea/blob/main"

# Timeout for external URL checks (seconds)
HTTP_TIMEOUT = 10

# ---------------------------------------------------------------------------
# Shared: collect doc files
# ---------------------------------------------------------------------------


def collect_doc_files() -> list[Path]:
    """Return all .md and .mdx files under DOCS_ROOT, skipping API ref."""
    files: list[Path] = []
    for ext in ("*.md", "*.mdx"):
        for p in sorted(DOCS_ROOT.rglob(ext)):
            rel = p.relative_to(DOCS_ROOT).as_posix()
            if any(rel.startswith(pfx) for pfx in SKIP_PREFIXES):
                continue
            files.append(p)
    return files


# ===================================================================
# LINK CHECKING
# ===================================================================

# Markdown link: [text](target) — but not images ![alt](src)
MD_LINK_RE = re.compile(r"(?<!!)\[(?:[^\]]*)\]\(([^)]+)\)")

# Mintlify Card href="..." (JSX)
HREF_RE = re.compile(r'href="([^"]+)"')


def extract_links(filepath: Path) -> list[tuple[int, str]]:
    """Return (line_number, raw_target) pairs from a file."""
    links: list[tuple[int, str]] = []
    text = filepath.read_text(encoding="utf-8", errors="replace")
    for lineno, line in enumerate(text.splitlines(), start=1):
        for m in MD_LINK_RE.finditer(line):
            links.append((lineno, m.group(1)))
        for m in HREF_RE.finditer(line):
            links.append((lineno, m.group(1)))
    return links


def is_external(target: str) -> bool:
    return target.startswith(("http://", "https://", "mailto:"))


def is_anchor_only(target: str) -> bool:
    return target.startswith("#")


def strip_anchor(target: str) -> str:
    return target.split("#", 1)[0]


def file_exists_mintlify(resolved: Path) -> bool:
    """Check whether the resolved target exists, trying Mintlify
    extension conventions (.md, .mdx, index files)."""
    if resolved.exists():
        return True
    if resolved.with_suffix(".md").exists():
        return True
    if resolved.with_suffix(".mdx").exists():
        return True
    if resolved.is_dir():
        if (resolved / "index.md").exists():
            return True
        if (resolved / "index.mdx").exists():
            return True
    return False


def check_external_url(url: str, cache: dict[str, int | str]) -> int | str:
    """HEAD-check an external URL.  Returns HTTP status code or error string.
    Results are cached for the session."""
    if url in cache:
        return cache[url]
    # Create an SSL context that doesn't verify (avoids cert issues in CI)
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    req = urllib.request.Request(
        url, method="HEAD", headers={"User-Agent": "mellea-doc-checker/1"}
    )
    try:
        with urllib.request.urlopen(req, timeout=HTTP_TIMEOUT, context=ctx) as resp:
            cache[url] = resp.status
            return resp.status
    except urllib.error.HTTPError as exc:
        cache[url] = exc.code
        return exc.code
    except Exception as exc:
        result = f"error: {exc}"
        cache[url] = result
        return result


def load_nav_pages() -> set[str]:
    """Return the set of page slugs declared in docs.json navigation."""
    docs_json = DOCS_ROOT / "docs.json"
    if not docs_json.exists():
        return set()
    with open(docs_json, encoding="utf-8") as f:
        data = json.load(f)
    pages: set[str] = set()

    def walk(node: object) -> None:
        if isinstance(node, str):
            pages.add(node)
        elif isinstance(node, list):
            for item in node:
                walk(item)
        elif isinstance(node, dict):
            for key in ("pages", "groups", "tabs"):
                if key in node:
                    walk(node[key])

    walk(data.get("navigation", {}))
    return pages


def load_navbar_links() -> list[tuple[str, str]]:
    """Return (label, href) for links in docs.json navbar."""
    docs_json = DOCS_ROOT / "docs.json"
    if not docs_json.exists():
        return []
    with open(docs_json, encoding="utf-8") as f:
        data = json.load(f)
    links: list[tuple[str, str]] = []
    navbar = data.get("navbar", {})
    primary = navbar.get("primary", {})
    if "href" in primary:
        links.append((primary.get("label", "primary"), primary["href"]))
    for item in navbar.get("links", []):
        if "href" in item:
            links.append((item.get("label", ""), item["href"]))
    return links


def run_link_checks(
    doc_files: list[Path], verbose: bool, check_external: bool
) -> tuple[list[str], list[str]]:
    """Return (errors, warnings) from link checking."""
    errors: list[str] = []
    warnings: list[str] = []
    url_cache: dict[str, int | str] = {}
    total_links = 0
    total_external = 0

    for filepath in doc_files:
        rel = filepath.relative_to(DOCS_ROOT)
        links = extract_links(filepath)

        for lineno, raw_target in links:
            total_links += 1

            # Pure anchor — skip
            if is_anchor_only(raw_target):
                if verbose:
                    print(f"  [skip] {rel}:{lineno} -> {raw_target} (anchor)")
                continue

            # External URL
            if is_external(raw_target):
                total_external += 1
                if check_external:
                    result = check_external_url(raw_target, url_cache)
                    if isinstance(result, int) and 200 <= result < 400:
                        if verbose:
                            print(f"  [ok]   {rel}:{lineno} -> {raw_target} ({result})")
                    elif isinstance(result, int) and result == 404:
                        errors.append(
                            f"  {rel}:{lineno} -> {raw_target}  [HTTP {result}]"
                        )
                    elif isinstance(result, int):
                        warnings.append(
                            f"  {rel}:{lineno} -> {raw_target}  [HTTP {result}]"
                        )
                    else:
                        warnings.append(f"  {rel}:{lineno} -> {raw_target}  [{result}]")
                elif verbose:
                    print(f"  [skip] {rel}:{lineno} -> {raw_target} (external)")
                continue

            # Internal link — resolve
            target_clean = strip_anchor(raw_target)
            if not target_clean:
                continue

            # Absolute Mintlify path
            if target_clean.startswith("/"):
                # Static assets
                if target_clean.startswith(("/images/", "/logo/")):
                    resolved = DOCS_ROOT / target_clean.lstrip("/")
                    if not resolved.exists():
                        errors.append(
                            f"  {rel}:{lineno} -> {raw_target}"
                            f"  [static asset not found]"
                        )
                    elif verbose:
                        print(f"  [ok]   {rel}:{lineno} -> {raw_target}")
                    continue

                resolved = DOCS_ROOT / target_clean.lstrip("/")
                if file_exists_mintlify(resolved):
                    if verbose:
                        print(f"  [ok]   {rel}:{lineno} -> {raw_target}")
                else:
                    errors.append(
                        f"  {rel}:{lineno} -> {raw_target}"
                        f"  [page not found under docs/docs/]"
                    )
                continue

            # Relative path
            source_dir = filepath.parent
            resolved = (source_dir / target_clean).resolve()

            # Check if the resolved path escapes DOCS_ROOT
            try:
                resolved.relative_to(DOCS_ROOT)
                inside_docs = True
            except ValueError:
                inside_docs = False

            if not inside_docs:
                # It might still exist in the repo...
                if resolved.exists() or Path(str(resolved)).exists():
                    # File exists in repo but won't work on Mintlify site
                    # Suggest the GitHub URL
                    try:
                        repo_rel = resolved.relative_to(REPO_ROOT)
                        suggested = f"{GITHUB_BASE}/{repo_rel}"
                    except ValueError:
                        suggested = "(could not compute GitHub URL)"
                    errors.append(
                        f"  {rel}:{lineno} -> {raw_target}"
                        f"  [escapes docs/ — won't work on Mintlify."
                        f" Suggest: {suggested}]"
                    )
                else:
                    errors.append(
                        f"  {rel}:{lineno} -> {raw_target}"
                        f"  [file not found, and escapes docs/]"
                    )
                if verbose:
                    print(f"  [ESC]  {rel}:{lineno} -> {raw_target}")
                continue

            # Normal internal link
            if file_exists_mintlify(resolved):
                if verbose:
                    print(f"  [ok]   {rel}:{lineno} -> {raw_target}")
            else:
                errors.append(f"  {rel}:{lineno} -> {raw_target}  [file not found]")

    # docs.json nav page slugs
    nav_pages = load_nav_pages()
    for slug in sorted(nav_pages):
        if any(slug.startswith(pfx) for pfx in SKIP_PREFIXES):
            continue
        resolved = DOCS_ROOT / slug
        if not file_exists_mintlify(resolved):
            errors.append(f"  docs.json nav: '{slug}' — file not found")
        elif verbose:
            print(f"  [ok]   docs.json nav: {slug}")

    # docs.json navbar links (external URLs)
    if check_external:
        for label, href in load_navbar_links():
            if is_external(href):
                result = check_external_url(href, url_cache)
                if isinstance(result, int) and result == 404:
                    errors.append(f"  docs.json navbar '{label}': {href}  [HTTP 404]")
                elif isinstance(result, int) and result >= 400:
                    warnings.append(
                        f"  docs.json navbar '{label}': {href}  [HTTP {result}]"
                    )
                elif isinstance(result, str):
                    warnings.append(f"  docs.json navbar '{label}': {href}  [{result}]")
                elif verbose:
                    print(f"  [ok]   docs.json navbar '{label}': {href}")

    print(
        f"\nLinks: scanned {len(doc_files)} files, "
        f"{total_links} links ({total_external} external)"
    )

    return errors, warnings


# ===================================================================
# CODE CHECKING
# ===================================================================

FENCE_OPEN_RE = re.compile(r"^```(?:python|py)\b.*$")
FENCE_CLOSE_RE = re.compile(r"^```\s*$")

# Known mellea names that should be imported when used
MELLEA_NAMES = {
    "mellea",
    "generative",
    "mify",
    "MelleaTool",
    "SimpleContext",
    "instruct",
    "start_session",
    "act",
    "aact",
    "GenSlot",
    "Requirement",
    "PydanticRequirement",
    "RegexRequirement",
    "ChatFormatter",
    "TemplateFormatter",
    "ModelOptions",
    "GuardianCheck",
    "MObject",
}


def extract_python_blocks(filepath: Path) -> list[tuple[int, str]]:
    """Return (start_line, code_text) for each Python fenced block."""
    blocks: list[tuple[int, str]] = []
    text = filepath.read_text(encoding="utf-8", errors="replace")
    lines = text.splitlines()
    in_block = False
    block_start = 0
    block_lines: list[str] = []

    for i, line in enumerate(lines):
        if not in_block:
            if FENCE_OPEN_RE.match(line.strip()):
                in_block = True
                block_start = i + 2  # 1-indexed, next line
                block_lines = []
        else:
            if FENCE_CLOSE_RE.match(line.strip()):
                in_block = False
                blocks.append((block_start, "\n".join(block_lines)))
            else:
                block_lines.append(line)
    return blocks


# SyntaxError messages that indicate a code *fragment* rather than a
# genuinely broken snippet.  These are downgraded to warnings.
_FRAGMENT_PATTERNS = (
    "'await' outside function",
    "'await' outside async function",
    "asynchronous comprehension outside of an asynchronous function",
    "unexpected indent",
    "'yield' outside function",
)


def check_syntax(code: str, filename: str) -> tuple[str | None, bool]:
    """Try to compile; return (error_message, is_fragment).

    is_fragment is True when the error is due to the snippet being an
    incomplete fragment (e.g. bare ``await`` or leading indentation)
    rather than genuinely broken syntax.
    """
    try:
        compile(code, filename, "exec")
        return None, False
    except SyntaxError as exc:
        detail = f"line {exc.lineno}: {exc.msg}" if exc.lineno else str(exc)
        msg = exc.msg or ""
        is_frag = any(pat in msg for pat in _FRAGMENT_PATTERNS)
        return f"SyntaxError: {detail}", is_frag


def extract_imports(code: str) -> list[tuple[str, int | None]]:
    """Return (module_name, lineno) for each import statement."""
    try:
        tree = ast.parse(code)
    except SyntaxError:
        return []
    imports: list[tuple[str, int | None]] = []
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                imports.append((alias.name, node.lineno))
        elif isinstance(node, ast.ImportFrom):
            if node.module:
                imports.append((node.module, node.lineno))
    return imports


def _mellea_module_exists(module_name: str) -> bool:
    """Check whether a mellea.* module exists on the filesystem.

    Walks from the repo's ``mellea/`` package directory, checking each
    dotted component resolves to a directory (package) or ``.py`` file.
    This avoids actually importing anything, so it's safe to call even
    when optional dependencies are missing.
    """
    mellea_pkg = REPO_ROOT / "mellea"
    if not mellea_pkg.is_dir():
        return False
    parts = module_name.split(".")
    current = REPO_ROOT
    for part in parts:
        candidate_dir = current / part
        candidate_file = current / f"{part}.py"
        if candidate_dir.is_dir():
            current = candidate_dir
        elif candidate_file.is_file():
            return True
        else:
            return False
    # Ended on a directory — valid package
    return (current / "__init__.py").is_file()


def module_importable(module_name: str) -> bool:
    """Check if module_name can be resolved without actually importing.

    For mellea.* modules, checks the full dotted path on the filesystem
    so that typos like ``mellea.stdlib.docs`` (should be
    ``mellea.stdlib.components.docs``) are caught even though the
    top-level ``mellea`` package exists.
    """
    if module_name.startswith("mellea"):
        return _mellea_module_exists(module_name)
    top = module_name.split(".")[0]
    try:
        return importlib.util.find_spec(top) is not None
    except (ModuleNotFoundError, ValueError):
        return False


def classify_module(name: str) -> str:
    if name.startswith("mellea"):
        return "mellea"
    if name.split(".")[0] in sys.stdlib_module_names:
        return "stdlib"
    return "third-party"


def check_missing_mellea_imports(code: str) -> list[str]:
    """Flag mellea names used but never imported."""
    try:
        tree = ast.parse(code)
    except SyntaxError:
        return []

    imported: set[str] = set()
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                imported.add(alias.asname or alias.name.split(".")[-1])
        elif isinstance(node, ast.ImportFrom):
            for alias in node.names:
                imported.add(alias.asname or alias.name)
            if node.module:
                imported.add(node.module.split(".")[0])

    used: set[str] = set()
    for node in ast.walk(tree):
        if isinstance(node, ast.Name):
            used.add(node.id)

    return sorted(used & MELLEA_NAMES - imported)


# Minimum lines for a code block to be considered for duplicate detection.
# Short snippets (imports, one-liners) are expected to repeat.
_DUPE_MIN_LINES = 4


def _code_hash(code: str) -> str:
    """Normalize and hash a code block for duplicate detection."""
    # Strip trailing whitespace per line, collapse blank lines
    normalized = "\n".join(line.rstrip() for line in code.splitlines()).strip()
    return hashlib.sha256(normalized.encode()).hexdigest()[:16]


def run_code_checks(
    doc_files: list[Path], verbose: bool
) -> tuple[list[str], list[str]]:
    """Return (errors, warnings) from code block checking."""
    errors: list[str] = []
    warnings: list[str] = []
    total_blocks = 0

    # Duplicate tracking: hash -> list of "file:line" labels
    seen_blocks: dict[str, list[str]] = {}

    for filepath in doc_files:
        rel = filepath.relative_to(DOCS_ROOT)
        blocks = extract_python_blocks(filepath)

        for start_line, code in blocks:
            total_blocks += 1
            label = f"{rel}:{start_line}"

            if verbose:
                preview = code.split("\n", 1)[0][:60]
                print(f"  [{total_blocks:3d}] {label}  {preview!r}")

            # Track duplicates (only for non-trivial blocks)
            line_count = len([ln for ln in code.splitlines() if ln.strip()])
            if line_count >= _DUPE_MIN_LINES:
                h = _code_hash(code)
                seen_blocks.setdefault(h, []).append(label)

            # 1. Syntax
            err, is_fragment = check_syntax(code, str(rel))
            if err:
                if is_fragment:
                    warnings.append(f"  {label} — {err} (fragment)")
                else:
                    errors.append(f"  {label} — {err}")
                continue

            # 2. Imports
            for mod_name, mod_line in extract_imports(code):
                cls = classify_module(mod_name)
                loc = f"{label}+{mod_line}" if mod_line else label
                if cls == "mellea" and not module_importable(mod_name):
                    warnings.append(
                        f"  {loc}: import {mod_name} — mellea submodule not found"
                    )
                elif cls == "third-party" and not module_importable(mod_name):
                    warnings.append(
                        f"  {loc}: import {mod_name}"
                        f" — third-party not installed (add install note?)"
                    )

            # 3. Missing mellea imports
            missing = check_missing_mellea_imports(code)
            if missing:
                warnings.append(
                    f"  {label}: uses {', '.join(missing)} without importing"
                )

    # 4. Duplicate code blocks (across different files)
    for h, locations in seen_blocks.items():
        if len(locations) < 2:
            continue
        # Only flag if the duplicates span different files
        files = {loc.rsplit(":", 1)[0] for loc in locations}
        if len(files) >= 2:
            locs = ", ".join(locations)
            warnings.append(
                f"  duplicate code block in {len(locations)} places: {locs}"
            )

    print(f"\nCode: scanned {len(doc_files)} files, {total_blocks} Python block(s)")

    return errors, warnings


# ===================================================================
# SHELL CHECKING
# ===================================================================

BASH_FENCE_RE = re.compile(r"^```(?:bash|shell|sh|zsh)\b.*$")

# Matches pip/uv install with [extras] — e.g. pip install mellea[litellm]
# Captures the full token including any surrounding quotes so we can check.
INSTALL_EXTRAS_RE = re.compile(
    r"""(?:pip|uv)\s+(?:install|pip\s+install)\s+  # pip install / uv install / uv pip install
        (?:(?:-\S+\s+)*)                            # optional flags like -U
        (['"]?)                                      # optional opening quote
        (\S+\[[^\]]+\])                              # package[extras]
        (['"]?)                                      # optional closing quote
    """,
    re.VERBOSE,
)


def extract_bash_blocks(filepath: Path) -> list[tuple[int, str]]:
    """Return (start_line, code_text) for each bash fenced block."""
    blocks: list[tuple[int, str]] = []
    text = filepath.read_text(encoding="utf-8", errors="replace")
    lines = text.splitlines()
    in_block = False
    block_start = 0
    block_lines: list[str] = []

    for i, line in enumerate(lines):
        if not in_block:
            if BASH_FENCE_RE.match(line.strip()):
                in_block = True
                block_start = i + 2
                block_lines = []
        else:
            if FENCE_CLOSE_RE.match(line.strip()):
                in_block = False
                blocks.append((block_start, "\n".join(block_lines)))
            else:
                block_lines.append(line)
    return blocks


def run_shell_checks(
    doc_files: list[Path], verbose: bool
) -> tuple[list[str], list[str]]:
    """Return (errors, warnings) from shell block checking."""
    errors: list[str] = []
    warnings: list[str] = []
    total_blocks = 0

    for filepath in doc_files:
        rel = filepath.relative_to(DOCS_ROOT)

        # Check bash code blocks
        blocks = extract_bash_blocks(filepath)
        for start_line, code in blocks:
            total_blocks += 1
            for i, line in enumerate(code.splitlines()):
                m = INSTALL_EXTRAS_RE.search(line)
                if m:
                    open_q, pkg, close_q = m.group(1), m.group(2), m.group(3)
                    quoted = open_q and close_q  # has matching quotes
                    if not quoted:
                        lineno = start_line + i
                        errors.append(
                            f"  {rel}:{lineno} — unquoted extras: {pkg}"
                            f'  [breaks in zsh — use "{pkg}"]'
                        )
                    elif verbose:
                        print(f"  [ok]   {rel}:{start_line + i} — quoted: {pkg}")

        # Also check inline code in markdown text for install commands
        text = filepath.read_text(encoding="utf-8", errors="replace")
        for lineno, line in enumerate(text.splitlines(), start=1):
            # Look for inline backtick commands: `pip install foo[bar]`
            for tick_m in re.finditer(r"`([^`]+)`", line):
                content = tick_m.group(1)
                extras_m = INSTALL_EXTRAS_RE.search(content)
                if extras_m:
                    open_q = extras_m.group(1)
                    pkg = extras_m.group(2)
                    close_q = extras_m.group(3)
                    quoted = open_q and close_q
                    if not quoted:
                        errors.append(
                            f"  {rel}:{lineno} — unquoted extras in"
                            f" inline code: {pkg}"
                            f'  [breaks in zsh — use "{pkg}"]'
                        )

    print(f"\nShell: scanned {len(doc_files)} files, {total_blocks} bash block(s)")

    return errors, warnings


# ===================================================================
# Main
# ===================================================================


def main() -> int:
    parser = argparse.ArgumentParser(
        description="Validate Mellea docs: links and code snippets"
    )
    all_checks = ["links", "code", "shell"]
    parser.add_argument(
        "checks",
        nargs="*",
        default=all_checks,
        metavar="CHECK",
        help="Which checks to run: links, code, shell (default: all)",
    )
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Show every item checked"
    )
    parser.add_argument(
        "--skip-external", action="store_true", help="Skip HTTP checks on external URLs"
    )
    args = parser.parse_args()

    if not DOCS_ROOT.is_dir():
        print(f"ERROR: docs root not found at {DOCS_ROOT}", file=sys.stderr)
        return 2

    doc_files = collect_doc_files()
    all_errors: list[str] = []
    all_warnings: list[str] = []

    if "links" in args.checks:
        print("=" * 60)
        print("LINK CHECKS")
        print("=" * 60)
        errs, warns = run_link_checks(
            doc_files, args.verbose, check_external=not args.skip_external
        )
        all_errors.extend(errs)
        all_warnings.extend(warns)

    if "code" in args.checks:
        print("\n" + "=" * 60)
        print("CODE CHECKS")
        print("=" * 60)
        errs, warns = run_code_checks(doc_files, args.verbose)
        all_errors.extend(errs)
        all_warnings.extend(warns)

    if "shell" in args.checks:
        print("\n" + "=" * 60)
        print("SHELL CHECKS")
        print("=" * 60)
        errs, warns = run_shell_checks(doc_files, args.verbose)
        all_errors.extend(errs)
        all_warnings.extend(warns)

    # Final summary
    print("\n" + "=" * 60)
    print("SUMMARY")
    print("=" * 60)

    if all_errors:
        print(f"\n{len(all_errors)} ERROR(s):\n")
        for e in all_errors:
            print(e)

    if all_warnings:
        print(f"\n{len(all_warnings)} WARNING(s):\n")
        for w in all_warnings:
            print(w)

    if not all_errors and not all_warnings:
        print("\nAll checks passed.")
    elif not all_errors:
        print(f"\nNo errors. {len(all_warnings)} warning(s) to review.")

    return 1 if all_errors else 0


if __name__ == "__main__":
    sys.exit(main())



"""The mellea standard library of components, sessions, and sampling strategies.

This package provides the high-level building blocks for writing generative programs
with mellea. It contains ready-to-use ``Component`` types (``Instruction``,
``Message``, ``Document``, ``Intrinsic``, ``SimpleComponent``, and more), context
implementations (``ChatContext``, ``SimpleContext``), sampling strategies (rejection
sampling, budget forcing), session management via ``MelleaSession``, and the
``@mify`` decorator for turning ordinary Python objects into components. Import from
the sub-packages — ``mellea.stdlib.components``, ``mellea.stdlib.sampling``, and
``mellea.stdlib.session`` — for day-to-day use.
"""



"""OpenTelemetry logging instrumentation for Mellea.

Provides log export using OpenTelemetry Logs API with OTLP exporter support.

Configuration via environment variables:
- MELLEA_LOGS_OTLP: Enable OTLP logs exporter (default: false)
- OTEL_EXPORTER_OTLP_LOGS_ENDPOINT: Logs-specific endpoint (optional, overrides general)
- OTEL_EXPORTER_OTLP_ENDPOINT: General endpoint for all signals (fallback)
- OTEL_SERVICE_NAME: Service name for logs (default: mellea)

Example:
    export MELLEA_LOGS_OTLP=true
    export OTEL_EXPORTER_OTLP_ENDPOINT=http://localhost:4317

Programmatic usage:
    from mellea.telemetry.logging import get_otlp_log_handler
    import logging

    logger = logging.getLogger("my_logger")
    handler = get_otlp_log_handler()
    if handler:
        logger.addHandler(handler)
        logger.info("This log will be exported via OTLP")
"""

import os
import warnings
from typing import Any

# Try to import OpenTelemetry, but make it optional
try:
    from opentelemetry._logs import set_logger_provider
    from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter
    from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
    from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
    from opentelemetry.sdk.resources import Resource

    _OTEL_AVAILABLE = True
except ImportError:
    _OTEL_AVAILABLE = False

# Configuration from environment variables
_LOGS_OTLP = _OTEL_AVAILABLE and os.getenv("MELLEA_LOGS_OTLP", "false").lower() in (
    "true",
    "1",
    "yes",
)
_OTLP_LOGS_ENDPOINT = os.getenv("OTEL_EXPORTER_OTLP_LOGS_ENDPOINT") or os.getenv(
    "OTEL_EXPORTER_OTLP_ENDPOINT"
)
_SERVICE_NAME = os.getenv("OTEL_SERVICE_NAME", "mellea")


def _setup_logger_provider() -> Any:
    """Set up the LoggerProvider with OTLP exporter.

    Returns:
        LoggerProvider instance or None if OpenTelemetry is not available or no endpoint configured
    """
    if not _OTEL_AVAILABLE:
        return None

    if not _OTLP_LOGS_ENDPOINT:
        warnings.warn(
            "OTLP logs exporter is enabled (MELLEA_LOGS_OTLP=true) but no endpoint is configured. "
            "Set OTEL_EXPORTER_OTLP_LOGS_ENDPOINT or OTEL_EXPORTER_OTLP_ENDPOINT to export logs.",
            UserWarning,
            stacklevel=2,
        )
        return None

    resource = Resource.create({"service.name": _SERVICE_NAME})  # type: ignore
    logger_provider = LoggerProvider(resource=resource)  # type: ignore

    try:
        otlp_exporter = OTLPLogExporter(endpoint=_OTLP_LOGS_ENDPOINT)  # type: ignore
        logger_provider.add_log_record_processor(
            BatchLogRecordProcessor(otlp_exporter)  # type: ignore
        )
    except Exception as e:
        warnings.warn(
            f"Failed to initialize OTLP logs exporter: {e}. "
            "Logs will not be exported via OTLP.",
            UserWarning,
            stacklevel=2,
        )
        return None

    set_logger_provider(logger_provider)  # type: ignore
    return logger_provider


# Initialize logger provider if OTLP logging is enabled
_logger_provider = None

if _LOGS_OTLP:
    _logger_provider = _setup_logger_provider()


def get_otlp_log_handler() -> Any:
    """Get an OTLP logging handler for Python's logging module.

    Returns:
        LoggingHandler instance if OTLP logging is enabled and configured,
        None otherwise.

    Example:
        import logging
        from mellea.telemetry.logging import get_otlp_log_handler

        logger = logging.getLogger("my_app")
        handler = get_otlp_log_handler()
        if handler:
            logger.addHandler(handler)
            logger.info("This log will be exported via OTLP")
    """
    if _logger_provider is None:
        return None

    try:
        handler = LoggingHandler(logger_provider=_logger_provider)  # type: ignore
        return handler
    except Exception as e:
        warnings.warn(
            f"Failed to create OTLP logging handler: {e}. "
            "Logs will not be exported via OTLP.",
            UserWarning,
            stacklevel=2,
        )
        return None



"""``Instruction`` component for instruct/validate/repair loops.

``Instruction`` is the primary component type used with ``MelleaSession.instruct``. It
packages a task ``description``, a list of ``Requirement`` constraints, optional
in-context-learning examples, a grounding context dict, user variables for Jinja2
template interpolation, and output/input prefix overrides into a single renderable
unit. The session's sampling strategy evaluates each requirement against the model's
output and may repair or resample until all requirements pass.
"""

from __future__ import annotations

from copy import deepcopy

import jinja2

from ...core import (
    CBlock,
    Component,
    ImageBlock,
    ModelOutputThunk,
    Requirement,
    TemplateRepresentation,
    blockify,
)
from ..requirements.requirement import reqify


class Instruction(Component[str]):
    """The Instruction in an instruct/validate/repair loop.

    Args:
        description (str | CBlock | None): The task description shown to the model.
        requirements (list[Requirement | str] | None): Constraints the output must satisfy.
        icl_examples (list[str | CBlock] | None): In-context-learning examples.
        grounding_context (dict[str, str | CBlock | Component] | None): Named context
            passages injected into the prompt.
        user_variables (dict[str, str] | None): Jinja2 variable substitutions applied
            to all string parameters.
        prefix (str | CBlock | None): A prefix prepended before the model's generation.
        output_prefix (str | CBlock | None): A prefix prepended to the model's output token
            stream (currently unsupported; must be ``None``).
        images (list[ImageBlock] | None): Images to include in the prompt.

    Attributes:
        requirements (list[Requirement]): The resolved list of requirement instances
            attached to this instruction.
    """

    def __init__(
        self,
        description: str | CBlock | None = None,
        requirements: list[Requirement | str] | None = None,
        icl_examples: list[str | CBlock] | None = None,
        grounding_context: dict[str, str | CBlock | Component] | None = None,
        user_variables: dict[str, str] | None = None,
        prefix: str | CBlock | None = None,
        output_prefix: str | CBlock | None = None,
        images: list[ImageBlock] | None = None,
    ):
        """Initialize Instruction, converting all string inputs to CBlocks and applying any Jinja2 variables."""
        requirements = [] if requirements is None else requirements
        icl_examples = [] if icl_examples is None else icl_examples
        grounding_context = dict() if grounding_context is None else grounding_context

        # Apply templates. All inputs must be strings if provided.
        if user_variables is not None:
            if description is not None:
                assert type(description) is str, (
                    "description must be a string when user_variables are provided"
                )
                description = Instruction.apply_user_dict_from_jinja(
                    user_variables, description
                )

            if prefix is not None:
                assert type(prefix) is str, (
                    "prefix must be a string when user_variables are provided"
                )
                prefix = Instruction.apply_user_dict_from_jinja(user_variables, prefix)

            # The following code would have to be un-commented-out if the assertion between this line and the code block is removed.
            assert output_prefix is None, (
                "output_prefix is not currently supported. The output_prefix serves as a prefix for the assistant's next message, and can be useful for 'priming' the model toward the right sort of answer. However, doing this requires using 'raw' endpoints instead of chat endpoints. Support for output_prefix will be re-enabled when we switch to span-first backend design."
            )
            # if output_prefix is not None:
            #     assert (
            #         type(output_prefix) == str
            #     ), "output prefix must be a string when user_variables are provided"
            #     output_prefix = Instruction.apply_user_dict_from_jinja(
            #         user_variables, output_prefix
            #     )

            for i, req in enumerate(requirements):
                assert type(req) is str or isinstance(req, Requirement), (
                    "requirements must be strings or Requirements when user_variables are provided"
                )
                if type(req) is str:
                    requirements[i] = Instruction.apply_user_dict_from_jinja(
                        user_variables, req
                    )
                elif isinstance(req, Requirement):
                    r = deepcopy(req)
                    r.description = Instruction.apply_user_dict_from_jinja(
                        user_variables,
                        req.description,  # type: ignore
                    )
                    requirements[i] = r

            for i, ex in enumerate(icl_examples):
                assert type(ex) is str, (
                    "icl_examples must be strings when user_variables are provided"
                )
                icl_examples[i] = Instruction.apply_user_dict_from_jinja(
                    user_variables, ex
                )

            for key in grounding_context:
                g = grounding_context[key]
                assert type(g) is str, (
                    "documents must be strings when user_variables are provided"
                )
                grounding_context[key] = Instruction.apply_user_dict_from_jinja(
                    user_variables, g
                )  # type: ignore

        self._description = blockify(description) if description is not None else None
        self._requirements: list[Requirement] = [reqify(r) for r in requirements]
        self._icl_examples: list[CBlock | Component] = [
            blockify(e) for e in icl_examples
        ]

        # Map all string values to CBlocks in the grounding context.
        self._grounding_context: dict[str, CBlock | Component] = {
            k: blockify(v) if isinstance(v, str) else v
            for k, v in grounding_context.items()
        }
        self._prefix = blockify(prefix) if prefix is not None else None
        self._output_prefix = (
            blockify(output_prefix) if output_prefix is not None else None
        )
        self._images = images
        self._repair_string: str | None = None

    def parts(self):
        """Returns all of the constituent parts of an Instruction."""
        # Add all of the optionally defined CBlocks/Components then filter Nones at the end.
        cs = [self._description, self._prefix, self._output_prefix]
        match self._grounding_context:
            case CBlock():
                cs.append(self._grounding_context)
            case _:
                cs.extend(list(self._grounding_context.values()))
        cs.extend(self._requirements)
        cs.extend(self._icl_examples)

        filtered: list[Component | CBlock] = list(filter(lambda x: x is not None, cs))  # type: ignore
        return filtered

    def format_for_llm(self) -> TemplateRepresentation:
        """Format this instruction for the language model.

        Returns:
            TemplateRepresentation: A template representation containing the
            description, requirements, in-context examples, grounding context,
            and optional prefix/repair fields.
        """
        return TemplateRepresentation(
            obj=self,
            args={
                "description": str(self._description),
                "requirements": [
                    r.description
                    for r in self._requirements
                    if r.description is not None
                    and r.description != ""
                    and not r.check_only
                ],
                "icl_examples": [str(e) for e in self._icl_examples],
                "grounding_context": self._grounding_context,
                "prefix": self._prefix if self._prefix is not None else None,
                "output_prefix": (
                    self._output_prefix if self._output_prefix is not None else None
                ),
                "repair": self._repair_string,
            },
            tools=None,
            images=self._images,
            template_order=["*", "Instruction"],
        )

    @staticmethod
    def apply_user_dict_from_jinja(user_dict: dict[str, str], s: str) -> str:
        """Render a Jinja2 template string using the provided variable dictionary.

        Args:
            user_dict (dict[str, str]): Mapping of Jinja2 variable names to their
                string replacement values.
            s (str): A string treated as a Jinja2 template to be rendered.

        Returns:
            str: The rendered string with all Jinja2 placeholders substituted.
        """
        assert s is not None
        return jinja2.Template(s).render(user_dict)

    @property
    def requirements(self) -> list[Requirement]:
        """Returns a list of Requirement instances."""
        return self._requirements

    def copy_and_repair(self, repair_string: str) -> Instruction:
        """Create a deep copy of this instruction with the repair string set.

        Args:
            repair_string (str): The repair feedback string to attach, typically
                describing which requirements failed and why.

        Returns:
            Instruction: A new ``Instruction`` identical to this one but with
            ``_repair_string`` set to ``repair_string``.
        """
        res = deepcopy(self)
        res._repair_string = repair_string
        return res

    def _parse(self, computed: ModelOutputThunk) -> str:
        """Parse the model output. Returns string value for now."""
        return computed.value if computed.value is not None else ""
